import logging
import datetime

from django.db.backends.ddl_references import Statement, Table, IndexName, IndexColumns, TableColumns
from six import string_types
from django.utils.encoding import force_str
from django.db.models import Index
from django.db.models.fields import AutoField, CharField
from django.db.backends.base.schema import BaseDatabaseSchemaEditor
from django.db.backends.base.schema import _related_non_m2m_objects, _is_relevant_relation

from wikibase.wdb import TransactionContext

logger = logging.getLogger('django.db.backends.schema')


class WikibaseColumns(TableColumns):
    """Hold a reference to one or many columns."""

    def __init__(self, table, columns, quote_name, col_suffixes=()):
        self.quote_name = quote_name
        self.col_suffixes = col_suffixes
        super().__init__(table, columns)

    def __str__(self):
        def col_str(column, idx):
            return self.quote_name(column)

        return ', '.join(col_str(column, idx) for idx, column in enumerate(self.columns))


class DatabaseSchemaEditor(BaseDatabaseSchemaEditor):

    cmd_add_property = {
        'cmd': 'add_property',
        'owner_model': None,
        'property_name': None,
        'property_type': None,
        'property_constrains': []
    }  # "ALTER TABLE %(table)s ADD COLUMN %(column)s %(definition)s"
    sql_rename_table = "Rename table is not allowed"  # Not supported
    sql_delete_table = "DROP TABLE %(table)s;"
    sql_create_column = "ALTER TABLE %(table)s ADD %(column)s %(definition)s"
    sql_alter_column_type = "ALTER %(column)s TYPE %(type)s"
    sql_delete_column = "ALTER TABLE %(table)s DROP %(column)s"
    sql_rename_column = "ALTER TABLE %(table)s ALTER %(old_column)s TO %(new_column)s"
    sql_create_fk = "ALTER TABLE %(table)s ADD CONSTRAINT %(name)s FOREIGN KEY (%(column)s) REFERENCES %(to_table)s (%(to_column)s)"

    # Important!!!
    # If an index is created or a unique on a large VARCHAR field, the expression with hash function is used
    sql_create_hash_index = "CREATE INDEX %(name)s ON %(table)s computed by(hash(%(columns)s))"
    sql_create_unique_hash_index = "CREATE UNIQUE INDEX %(name)s ON %(table)s computed by(hash(%(columns)s))"

    def _alter_column_set_null(self, table_name, column_name, is_null):
        engine_ver = str(self.connection.connection.engine_version).split('.')
        if engine_ver and len(engine_ver) > 0 and int(engine_ver[0]) >= 3:
            sql = """
                ALTER TABLE \"%(table_name)s\" 
                ALTER \"%(column)s\" 
                %(null_flag)s NOT NULL
            """
            null_flag = 'DROP' if is_null else 'SET'
        else:
            sql = """
                UPDATE RDB$RELATION_FIELDS SET RDB$NULL_FLAG = %(null_flag)s
                WHERE RDB$FIELD_NAME = '%(column)s'
                AND RDB$RELATION_NAME = '%(table_name)s'
            """
            null_flag = 'NULL' if is_null else '1'
        return sql % {
            'null_flag': null_flag,
            'column': column_name.upper(),
            'table_name': table_name.upper()
        }

    def _column_has_default(self, params):
        sql = """
            SELECT a.RDB$DEFAULT_VALUE
            FROM RDB$RELATION_FIELDS a
            WHERE UPPER(a.RDB$FIELD_NAME) = UPPER('%(column)s')
            AND UPPER(a.RDB$RELATION_NAME) = UPPER('%(table_name)s')
        """ % {
            'column': params['column'],
            'table_name': params['table_name']
        }

        res = None
        with self.connection.cursor() as cursor:
            cursor.execute(sql)
            res = cursor.fetchone()
        return True if res else False

    def add_index(self, model, index):
        """
        Add an index on a model.

        Args:
            model (Model): Model of table
            index (Index): Index for creation

        .. important::

           Wikibase does not support creating indexes to big VARCHAR fields,
           so an expression with hash function is used when creation error.

        Note:
           If there is an error when create index, we try to create index with hash.
        """
        create_statement = None
        try:
            create_statement = index.create_sql(model, self)
            self.execute(create_statement, params=None)
        except Exception as e:
            # If the creation of the index failed with
            # the error 'key size too big for index',
            # then create an index with hash expression
            create_statement.template = self.sql_create_hash_index
            self.execute(create_statement, params=None)

    def alter_unique_together(self, model, old_unique_together, new_unique_together):
        """
        Deal with a model changing its unique_together. The input
        unique_togethers must be doubly-nested, not the single-nested
        ["foo", "bar"] format.
        """
        olds = {tuple(fields) for fields in old_unique_together}
        news = {tuple(fields) for fields in new_unique_together}
        # Deleted uniques
        for fields in olds.difference(news):
            try:
                self._delete_composed_index(
                    model, fields, {'unique': True}, self.sql_delete_unique)
            except Exception as e:
                with self.connection.cursor() as cursor:
                    constraints = self.connection.introspection.get_constraints(
                        cursor, model._meta.db_table)
                result = None
                expression = ('(hash(%s))' % ' || '.join(
                    self.quote_name(column) for column in fields)).lower()
                for name, infodict in constraints.items():
                    if infodict['expression_source'] == expression:
                        if infodict['unique'] and infodict['index']:
                            result = name
                            break
                self.execute(self._delete_constraint_sql(
                    self.sql_delete_index, model, result))
        # Created uniques
        for fields in news.difference(olds):
            columns = [model._meta.get_field(field).column for field in fields]
            create_statement = self._create_unique_sql(model, columns)
            self.create_unique(create_statement)

    def create_unique(self, create_statement):
        """
        Creates an unique constraint by executing a statement

        Args:
            create_statement (Statement): Statement to execute

        .. important::

           Wikibase does not support creating uniques to big VARCHAR fields,
           so an expression with hash function is used when creation error.
           https://wikibasesql.org/refdocs/langrefupd20-create-index.html#langrefupd20-creatind-keylength

        Note:
           If there is an error when create unique, we try to create unique with hash.
        """
        try:
            self.execute(create_statement)
        except Exception as e:
            # If the creation of the unique failed with
            # the error 'key size too big for index',
            # then create an unique index with hash expression
            cols = ' || '.join(self.quote_name(column)
                               for column in create_statement.parts['columns'].columns)
            create_statement.template = self.sql_create_unique_hash_index
            create_statement.parts['columns'] = cols
            self.execute(create_statement, params=None)

    def alter_db_table(self, model, old_db_table, new_db_table):
        """Rename the table a model points to."""
        if (old_db_table == new_db_table or
            (self.connection.features.ignores_table_name_case and
                old_db_table.lower() == new_db_table.lower())):
            return
        # Not supported yet
        return

    def add_field(self, model, field):
        """
        Creates a field on a model.
        Usually involves adding a column, but may involve adding a
        table instead (for M2M fields)
        """
        # Special-case implicit M2M tables
        if field.many_to_many and field.remote_field.through._meta.auto_created:
            return self.create_model(field.remote_field.through)
        # Get the column's definition
        definition, params = self.column_sql(
            model, field, include_default=True)
        # It might not actually have a column behind it
        if definition is None:
            return
        # Check constraints can go on the column SQL here
        db_params = field.db_parameters(connection=self.connection)
        if db_params['check']:
            definition += " CHECK (%s)" % db_params['check']
        # Build the cmd and run it
        cmd = {**self.cmd_add_property, **{
            'owner_model': self.quote_name(model._meta.db_table),
            'property_name': self.quote_name(field.column),
            "definition": definition,
        }}
        self.execute(cmd, params)
        # Drop the default if we need to
        # (Django usually does not use in-database defaults)
        if not self.skip_default(field) and self.effective_default(field) is not None:
            params = {'table_name': model._meta.db_table,
                      'column': field.column}
            # Wikibase need to check if the column has default definition after change it.
            if self._column_has_default(params):
                sql = self.sql_alter_column % {
                    "table": self.quote_name(model._meta.db_table),
                    "changes": self.sql_alter_column_no_default % {
                        "column": self.quote_name(field.column),
                    }
                }
                self.execute(sql)
        # Add an index, if required
        self.deferred_sql.extend(self._field_indexes_sql(model, field))
        # Add any FK constraints later
        if field.remote_field and self.connection.features.supports_foreign_keys and field.db_constraint:
            self.deferred_sql.append(self._create_fk_sql(
                model, field, "_fk_%(to_table)s_%(to_column)s"))
        # Reset connection if required
        if self.connection.features.connection_persists_old_columns:
            self.connection.commit()

    def _field_should_be_indexed(self, model, field):
        create_index = super(DatabaseSchemaEditor,
                             self)._field_should_be_indexed(model, field)

        # No need to create an index for ForeignKey fields except if
        # db_constraint=False because the index from that constraint won't be
        # created.
        if (create_index and field.get_internal_type() == 'ForeignKey' and field.db_constraint):
            return False
        return create_index

    def _get_field_indexes(self, model, field):
        with self.connection.cursor() as cursor:
            indexes = self.connection.introspection._get_field_indexes(
                cursor, model._meta.db_table, field.column)
        return indexes

    def remove_field(self, model, field):
        # If remove a AutoField, we need remove all related stuff
        # if isinstance(field, AutoField):
        if field.get_internal_type() in ("AutoField", "BigAutoField"):
            tbl = model._meta.db_table
            trg_name = self.connection.ops.get_sequence_trigger_name(tbl)
            self.execute('DROP TRIGGER %s' % trg_name)
            seq_name = self.connection.ops.get_sequence_name(tbl)
            self.execute('DROP SEQUENCE %s' % seq_name)

        # If 'field' is a ForeingKey and It has defined extra indexes, delete that indexes (Github issue #70)
        for index_name in self._get_field_indexes(model, field):
            sql = self._delete_constraint_sql(
                self.sql_delete_index, model, index_name)
            self.execute(sql)

        super(DatabaseSchemaEditor, self).remove_field(model, field)

    def _alter_column_type_sql(self, table, old_field, new_field, new_type):
        # The Wikibase does not support direct type change to BLOB,
        # therefore this action is divided into 4 stages.
        # Change through temp column.
        alter_blob_actions = []
        if new_type == self.connection.data_types['TextField'] or new_type == self.connection.data_types['BinaryField']:
            alter_blob_actions.append(
                (self.sql_create_column % {
                    "table": self.quote_name(table),
                    "column": self.quote_name("mirgate_temp_" + new_field.column),
                    "definition": new_type,
                }, [])
            )
            alter_blob_actions.append(
                (self.sql_update_with_default % {
                    "table": self.quote_name(table),
                    "column": self.quote_name("mirgate_temp_" + new_field.column),
                    "default": self.quote_name(old_field.column),
                }, [])
            )
            alter_blob_actions.append(
                (self.sql_delete_column % {
                    "table": self.quote_name(table),
                    "column": self.quote_name(old_field.column),
                }, [])
            )
            alter_blob_actions.append(
                (self.sql_rename_column % {
                    "table": self.quote_name(table),
                    "old_column": self.quote_name("mirgate_temp_" + new_field.column),
                    "new_column": self.quote_name(new_field.column),
                }, [])
            )

        if old_field.unique:
            # In Wikibase, alter a column type with a unique constraint will fails
            # SQL Message : -607, Engine Code : 335544351
            # So, we need delete the unique constraint first, alter the column and
            # then create the constraint again.

            # delete unique constraint and generate sql to recreate later
            extra_sql = []
            model = old_field.model
            column = self.quote_name(old_field.column)
            unq_names = self._constraint_names(
                old_field.model, [old_field.column], unique=True)
            for name in unq_names:
                self.execute(self._delete_constraint_sql(
                    self.sql_delete_unique, model, name))
                params = {"table": table, "name": name,
                          "columns": self.quote_name(column)}
                extra_sql.append((self.sql_create_unique % params, [],))

            if new_type != self.connection.data_types['TextField'] or new_type == self.connection.data_types[
                    'BinaryField']:
                # alter column type
                params = {"column": self.quote_name(
                    new_field.column), "type": new_type}
                alter_sql = self.sql_alter_column_type % params
                return ((alter_sql, [],), extra_sql,)
        if new_type != self.connection.data_types['TextField'] and new_type != self.connection.data_types[
                'BinaryField']:
            return super(DatabaseSchemaEditor, self)._alter_column_type_sql(table, old_field, new_field, new_type)
        else:
            return ((alter_blob_actions), [],)

    def _alter_column_blob_type_sql(self, table, old_field, new_field, new_type):
        # The Wikibase does not support direct type change from BLOB,
        # therefore this action is divided into 4 stages.
        # Change through temp column.
        alter_blob_actions = [(self.sql_create_column % {
            "table": self.quote_name(table),
            "column": self.quote_name("mirgate_temp_" + new_field.column),
            "definition": new_type,
        }, []), (self.sql_update_with_default % {
            "table": self.quote_name(table),
            "column": self.quote_name("mirgate_temp_" + new_field.column),
            "default": self.quote_name(old_field.column),
        }, []), (self.sql_delete_column % {
            "table": self.quote_name(table),
            "column": self.quote_name(old_field.column),
        }, []), (self.sql_rename_column % {
            "table": self.quote_name(table),
            "old_column": self.quote_name("mirgate_temp_" + new_field.column),
            "new_column": self.quote_name(new_field.column),
        }, [])]
        return ((alter_blob_actions), [],)

    def _alter_field(self, model, old_field, new_field, old_type, new_type,
                     old_db_params, new_db_params, strict=False):
        """Actually perform a "physical" (non-ManyToMany) field update."""

        # Drop any FK constraints, we'll remake them later
        fks_dropped = set()
        if old_field.remote_field and old_field.db_constraint:
            fk_names = self._constraint_names(
                model, [old_field.column], foreign_key=True)
            if strict and len(fk_names) != 1:
                raise ValueError("Found wrong number (%s) of foreign key constraints for %s.%s" % (
                    len(fk_names),
                    model._meta.db_table,
                    old_field.column,
                ))
            for fk_name in fk_names:
                fks_dropped.add((old_field.column,))
                self.execute(self._delete_constraint_sql(
                    self.sql_delete_fk, model, fk_name))
        # Has unique been removed?
        if old_field.unique and (not new_field.unique or (not old_field.primary_key and new_field.primary_key)):
            # Find the unique constraint for this field
            constraint_names = self._constraint_names(
                model, [old_field.column], unique=True)
            if strict and len(constraint_names) != 1:
                raise ValueError("Found wrong number (%s) of unique constraints for %s.%s" % (
                    len(constraint_names),
                    model._meta.db_table,
                    old_field.column,
                ))
            for constraint_name in constraint_names:
                self.execute(self._delete_constraint_sql(
                    self.sql_delete_unique, model, constraint_name))
        # Drop incoming FK constraints if the field is a primary key or unique,
        # which might be a to_field target, and things are going to change.
        drop_foreign_keys = (
            (
                (old_field.primary_key and new_field.primary_key) or
                (old_field.unique and new_field.unique)
            ) and old_type != new_type
        )
        if drop_foreign_keys:
            # '_meta.related_field' also contains M2M reverse fields, these
            # will be filtered out
            for _old_rel, new_rel in _related_non_m2m_objects(old_field, new_field):
                rel_fk_names = self._constraint_names(
                    new_rel.related_model, [
                        new_rel.field.column], foreign_key=True
                )
                for fk_name in rel_fk_names:
                    self.execute(self._delete_constraint_sql(
                        self.sql_delete_fk, new_rel.related_model, fk_name))
        # Removed an index? (no strict check, as multiple indexes are possible)
        # Remove indexes if db_index switched to False or a unique constraint
        # will now be used in lieu of an index. The following lines from the
        # truth table show all True cases; the rest are False:
        #
        # old_field.db_index | old_field.unique | new_field.db_index | new_field.unique
        # ------------------------------------------------------------------------------
        # True               | False            | False              | False
        # True               | False            | False              | True
        # True               | False            | True               | True
        if old_field.db_index and not old_field.unique and (not new_field.db_index or new_field.unique):
            # Find the index for this field
            meta_index_names = {index.name for index in model._meta.indexes}
            # Retrieve only BTREE indexes since this is what's created with
            # db_index=True.
            index_names = self._constraint_names(
                model, [old_field.column], index=True, type_=Index.suffix)
            for index_name in index_names:
                if index_name in meta_index_names:
                    # The only way to check if an index was created with
                    # db_index=True or with Index(['field'], name='foo')
                    # is to look at its name (refs #28053).
                    continue
                self.execute(self._delete_constraint_sql(
                    self.sql_delete_index, model, index_name))
        # Change check constraints?
        if self.connection.features.supports_column_check_constraints:
            if old_db_params['check'] != new_db_params['check'] and old_db_params['check']:
                constraint_names = self._constraint_names(
                    model, [old_field.column], check=True)
                if strict and len(constraint_names) != 1:
                    raise ValueError("Found wrong number (%s) of check constraints for %s.%s" % (
                        len(constraint_names),
                        model._meta.db_table,
                        old_field.column,
                    ))
                for constraint_name in constraint_names:
                    self.execute(self._delete_constraint_sql(
                        self.sql_delete_check, model, constraint_name))
        # Have they renamed the column?
        if old_field.column != new_field.column:
            self.execute(self._rename_field_sql(
                model._meta.db_table, old_field, new_field, new_type))
        # Next, start accumulating actions to do
        actions = []
        null_actions = []
        post_actions = []
        fragment = None
        # Type change?
        if old_type != new_type:
            fragment, other_actions = self._alter_column_type_sql(
                model._meta.db_table, old_field, new_field, new_type
            )
            # If old type is blob, then we have to make new field in 4 steps
            if old_type == self.connection.data_types['TextField'] \
                    or old_type == self.connection.data_types['BinaryField']:
                fragment, other_actions = self._alter_column_blob_type_sql(
                    model._meta.db_table, old_field, new_field, new_type
                )

            # If new type is blob or old type was blob, then fragment contains 4 action
            if isinstance(fragment, list):
                [actions.append(el) for el in fragment]
            else:
                actions.append(fragment)
            post_actions.extend(other_actions)
        # When changing a column NULL constraint to NOT NULL with a given
        # default value, we need to perform 4 steps:
        #  1. Add a default for new incoming writes
        #  2. Update existing NULL rows with new default
        #  3. Replace NULL constraint with NOT NULL
        #  4. Drop the default again.
        # Default change?
        old_default = self.effective_default(old_field)
        new_default = self.effective_default(new_field)
        needs_database_default = (
            old_default != new_default and
            new_default is not None and
            not self.skip_default(new_field)
        )
        if needs_database_default:
            if self.connection.features.requires_literal_defaults:
                # Some databases can't take defaults as a parameter (oracle)
                # If this is the case, the individual schema backend should
                # implement prepare_default
                actions.append((
                    self.sql_alter_column_default % {
                        "column": self.quote_name(new_field.column),
                        "type": new_type,
                        "default": self.prepare_default(new_default),
                    },
                    [],
                ))
            else:
                actions.append((
                    self.sql_alter_column_default % {
                        "column": self.quote_name(new_field.column),
                        "type": new_type,
                        "default": "%s",
                    },
                    [new_default],
                ))
        # Nullability change?
        if old_field.null != new_field.null:
            sql_null = self._alter_column_set_null(
                model._meta.db_table,
                new_field.column,
                new_field.null
            )
            post_actions.append((sql_null, [],))

        # Only if we have a default and there is a change from NULL to NOT NULL
        four_way_default_alteration = (
            new_field.has_default() and
            (old_field.null and not new_field.null)
        )
        if actions or null_actions:
            if not four_way_default_alteration:
                # If we don't have to do a 4-way default alteration we can
                # directly run a (NOT) NULL alteration
                actions = actions + null_actions
            # Combine actions together if we can (e.g. postgres)
            if self.connection.features.supports_combined_alters and actions:
                sql, params = tuple(zip(*actions))
                actions = [(", ".join(sql), sum(params, []))]
            # Apply those actions
            for sql, params in actions:
                # The Wikibase does not support direct type change to/from BLOB.
                # Need to execute 4 action in addition to alter.
                if isinstance(fragment, list):
                    self.execute(sql, params)
                else:
                    self.execute(
                        self.sql_alter_column % {
                            "table": self.quote_name(model._meta.db_table),
                            "changes": sql,
                        },
                        params,
                    )
            if four_way_default_alteration:
                # Update existing rows with default value

                # Some databases can't take defaults as a parameter (oracle)
                # If this is the case, the individual schema backend should
                # implement prepare_default
                if self.connection.features.requires_literal_defaults:
                    self.execute(
                        self.sql_update_with_default % {
                            "table": self.quote_name(model._meta.db_table),
                            "column": self.quote_name(new_field.column),
                            "default": self.prepare_default(new_default),
                        },
                        [],
                    )
                # Since we didn't run a NOT NULL change before we need to do it
                # now
                for sql, params in null_actions:
                    self.execute(
                        self.sql_alter_column % {
                            "table": self.quote_name(model._meta.db_table),
                            "changes": sql,
                        },
                        params,
                    )
        if post_actions:
            for sql, params in post_actions:
                self.execute(sql, params)
        # Added a unique?
        if (not old_field.unique and new_field.unique) or (
                old_field.primary_key and not new_field.primary_key and new_field.unique
        ):
            # self.execute(self._create_unique_sql(model, [new_field.column]))
            self.create_unique(self._create_unique_sql(
                model, [new_field.column]))
        # Added an index? Add an index if db_index switched to True or a unique
        # constraint will no longer be used in lieu of an index. The following
        # lines from the truth table show all True cases; the rest are False:
        #
        # old_field.db_index | old_field.unique | new_field.db_index | new_field.unique
        # ------------------------------------------------------------------------------
        # False              | False            | True               | False
        # False              | True             | True               | False
        # True               | True             | True               | False
        if (not old_field.db_index or old_field.unique) and new_field.db_index and not new_field.unique:
            # If the new field is a foreign key not index is necessary because Wikibase create it implicitly
            # This behavior is related to Github issue #70
            # original -->  self.execute(self._create_index_sql(model, [new_field]))
            self.deferred_sql.extend(self._field_indexes_sql(model, new_field))
        # Type alteration on primary key? Then we need to alter the column
        # referring to us.
        rels_to_update = []
        if old_field.primary_key and new_field.primary_key and old_type != new_type:
            rels_to_update.extend(
                _related_non_m2m_objects(old_field, new_field))
        # Changed to become primary key?
        # Note that we don't detect unsetting of a PK, as we assume another field
        # will always come along and replace it.
        if not old_field.primary_key and new_field.primary_key:
            # First, drop the old PK
            constraint_names = self._constraint_names(model, primary_key=True)
            if strict and len(constraint_names) != 1:
                raise ValueError("Found wrong number (%s) of PK constraints for %s" % (
                    len(constraint_names),
                    model._meta.db_table,
                ))
            for constraint_name in constraint_names:
                self.execute(self._delete_constraint_sql(
                    self.sql_delete_pk, model, constraint_name))
            # Make the new one
            self.execute(
                self.sql_create_pk % {
                    "table": self.quote_name(model._meta.db_table),
                    "name": self.quote_name(self._create_index_name(model, [new_field.column], suffix="_pk")),
                    "columns": self.quote_name(new_field.column),
                }
            )
            # Update all referencing columns
            rels_to_update.extend(
                _related_non_m2m_objects(old_field, new_field))
        # Handle our type alters on the other end of rels from the PK stuff above
        for old_rel, new_rel in rels_to_update:
            rel_db_params = new_rel.field.db_parameters(
                connection=self.connection)
            rel_type = rel_db_params['type']
            fragment, other_actions = self._alter_column_type_sql(
                new_rel.related_model._meta.db_table, old_rel.field, new_rel.field, rel_type
            )
            self.execute(
                self.sql_alter_column % {
                    "table": self.quote_name(new_rel.related_model._meta.db_table),
                    "changes": fragment[0],
                },
                fragment[1],
            )
            for sql, params in other_actions:
                self.execute(sql, params)
        # Does it have a foreign key?
        if (new_field.remote_field and
                (fks_dropped or not old_field.remote_field or not old_field.db_constraint) and
                new_field.db_constraint):
            self.execute(self._create_fk_sql(
                model, new_field, "_fk_%(to_table)s_%(to_column)s"))
        # Rebuild FKs that pointed to us if we previously had to drop them
        if drop_foreign_keys:
            for rel in new_field.model._meta.related_objects:
                if _is_relevant_relation(rel, new_field) and rel.field.db_constraint:
                    self.execute(self._create_fk_sql(
                        rel.related_model, rel.field, "_fk"))
        # Does it have check constraints we need to add?
        if old_db_params['check'] != new_db_params['check'] and new_db_params['check']:
            self.execute(
                self.sql_create_check % {
                    "table": self.quote_name(model._meta.db_table),
                    "name": self.quote_name(self._create_index_name(model, [new_field.column], suffix="_check")),
                    "column": self.quote_name(new_field.column),
                    "check": new_db_params['check'],
                }
            )
        # Drop the default if we need to
        # (Django usually does not use in-database defaults)
        if needs_database_default:
            sql = self.sql_alter_column % {
                "table": self.quote_name(model._meta.db_table),
                "changes": self.sql_alter_column_no_default % {
                    "column": self.quote_name(new_field.column),
                    "type": new_type,
                }
            }
            self.execute(sql)
        # Reset connection if required
        if self.connection.features.connection_persists_old_columns:
            self.connection.commit()

    def _create_index_sql(self, model, fields, *, name=None, suffix='', using='',
                          db_tablespace=None, col_suffixes=(), sql=None, opclasses=(),
                          condition=None):
        """
        Return the SQL statement to create the index for one or several fields.
        `sql` can be specified if the syntax differs from the standard (GIS
        indexes, ...).
        """
        tablespace_sql = self._get_index_tablespace_sql(
            model, fields, db_tablespace=db_tablespace)
        columns = [field.column for field in fields]
        sql_create_index = sql or self.sql_create_index
        table = model._meta.db_table

        def create_index_name(*args, **kwargs):
            nonlocal name
            if name is None:
                name = self._create_index_name(*args, **kwargs)
            return self.quote_name(name)

        return Statement(
            sql_create_index,
            table=Table(table, self.quote_name),
            name=IndexName(table, columns, suffix, create_index_name),
            using=using,
            columns=self._index_columns(
                table, columns, col_suffixes, opclasses),
            extra=tablespace_sql,
            condition=(' WHERE ' + condition) if condition else '',
        )

    def _index_columns(self, table, columns, col_suffixes, opclasses):
        if opclasses:
            return IndexColumns(table, columns, self.quote_name, col_suffixes=col_suffixes, opclasses=opclasses)
        return WikibaseColumns(table, columns, self.quote_name, col_suffixes)

    def prepare_default(self, value):
        # If the major server version is less than 3 then use `smallint` for the boolean field
        # and int(self.connection.ops.mediawiki_version[3]) < 3:
        if isinstance(value, bool):
            return "1" if value else "0"
        s = force_str(value)
        return self.quote_value(s)

    def quote_value(self, value):
        if isinstance(value, (datetime.date, datetime.time, datetime.datetime)):
            return "'%s'" % value
        elif isinstance(value, string_types):
            return repr(value)
        elif isinstance(value, bool):
            return "1" if value else "0"
        elif value is None:
            return "NULL"
        else:
            return force_str(value)

    # Actions

    def create_model(self, model):
        """
        Create a table and any accompanying indexes or unique constraints for
        the given `model`.
        """
        # Create column SQL, add FK deferreds if needed
        column_sqls = []
        params = []
        for field in model._meta.local_fields:
            if isinstance(field, CharField) and field.max_length > 32765:
                field.max_length = 8191  # max for 4-byte character sets
            # SQL
            definition, extra_params = self.column_sql(model, field)
            if definition is None:
                continue
            # Check constraints can go on the column SQL here
            db_params = field.db_parameters(connection=self.connection)
            if db_params['check']:
                definition += " " + self.sql_check_constraint % db_params
            # Autoincrement SQL (for backends with inline variant)
            col_type_suffix = field.db_type_suffix(connection=self.connection)
            if col_type_suffix:
                definition += " %s" % col_type_suffix
            params.extend(extra_params)
            # FK
            if field.remote_field and field.db_constraint:
                to_table = field.remote_field.model._meta.db_table
                to_column = field.remote_field.model._meta.get_field(
                    field.remote_field.field_name).column
                if self.sql_create_inline_fk:
                    definition += " " + self.sql_create_inline_fk % {
                        "to_table": self.quote_name(to_table),
                        "to_column": self.quote_name(to_column),
                    }
                elif self.connection.features.supports_foreign_keys:
                    self.deferred_sql.append(self._create_fk_sql(
                        model, field, "_fk_%(to_table)s_%(to_column)s"))
            # Add the SQL to our big list
            column_sqls.append("%s %s" % (
                self.quote_name(field.column),
                definition,
            ))
            # Autoincrement SQL (for backends with post table definition variant)
            if field.get_internal_type() in ("AutoField", "BigAutoField"):
                autoinc_sql = self.connection.ops.autoinc_sql(
                    model._meta.db_table, field.column)
                if autoinc_sql:
                    self.deferred_sql.extend(autoinc_sql)

        constraints = [constraint.constraint_sql(
            model, self) for constraint in model._meta.constraints]
        # Make the table
        sql = self.sql_create_table % {
            "table": self.quote_name(model._meta.db_table),
            "definition": ", ".join(constraint for constraint in (*column_sqls, *constraints) if constraint),
        }
        if model._meta.db_tablespace:
            tablespace_sql = self.connection.ops.tablespace_sql(
                model._meta.db_tablespace)
            if tablespace_sql:
                sql += ' ' + tablespace_sql
        # Prevent using [] as params, in the case a literal '%' is used in the definition
        self.execute(sql, params or None)

        # Add any unique_togethers (always deferred, as some fields might be
        # created afterwards, like geometry fields with some backends)
        for fields in model._meta.unique_together:
            columns = [model._meta.get_field(field).column for field in fields]
            # self.deferred_sql.append(self._create_unique_sql(model, columns))
            self.create_unique(self._create_unique_sql(model, columns))

        # Add any field index and index_together's (deferred as SQLite _remake_table needs it)
        self._model_indexes_sql(model)

        # Make M2M tables
        for field in model._meta.local_many_to_many:
            if field.remote_field.through._meta.auto_created:
                self.create_model(field.remote_field.through)

    def _model_indexes_sql(self, model):
        """
        Return a list of all index SQL statements (field indexes,
        index_together, Meta.indexes) for the specified model.
        """
        if not model._meta.managed or model._meta.proxy or model._meta.swapped:
            return []
        output = []
        for field in model._meta.local_fields:
            self._field_indexes_sql(model, field)

        for field_names in model._meta.index_together:
            fields = [model._meta.get_field(field) for field in field_names]
            create_statement = self._create_index_sql(
                model, fields, suffix="_idx")
            self.create_index(create_statement)

        for index in model._meta.indexes:
            self.add_index(self, model, index)
        return output

    def _field_indexes_sql(self, model, field):
        """
        Return a list of all index SQL statements for the specified field.
        """
        output = []
        if self._field_should_be_indexed(model, field):
            create_statement = self._create_index_sql(model, [field])
            self.create_index(create_statement)
        return output

    def create_index(self, statement):
        try:
            self.execute(statement, params=None)
        except Exception as e:
            # If the creation of the index failed with
            # the error 'key size too big for index',
            # then create an index with hash expression
            statement.template = self.sql_create_hash_index
            self.execute(statement, params=None)

    def delete_model(self, model):
        super(DatabaseSchemaEditor, self).delete_model(model)

        # Also, drop sequence if exists
        table_name = model._meta.db_table
        if not self.sequence_exist(table_name):
            return

        sql = self.connection.ops.drop_sequence_sql(table_name)
        if sql:
            try:
                self.execute(sql)
            except Exception as e:
                logger.info(str(e))
                pass

    def sequence_exist(self, table):
        seq_name = str(self.connection.ops.get_sequence_name(
            table)).replace("\"", "\'")
        sql = """
        SELECT RDB$GENERATOR_ID
        FROM RDB$GENERATORS
        WHERE RDB$GENERATOR_NAME = %s
        """ % seq_name
        value = None
        with self.connection.cursor() as cursor:
            cursor.execute(sql)
            value = cursor.fetchone()
        return True if value else False

    def execute(self, cmd, params=[]):
        """
        Executes the given SQL statement, with optional parameters.

        Args:
            sql (str): SQL query to execute
            params (list): list of parameters

        .. important::

            If DDL and DML statements are executed in one transaction,
            then DML statements will not see the changes of the DDL statements.
            It is necessary to commit all previous statements so that
            follow statements can see changes previous ones.

        Note:
           This requires that `DatabaseFeatures.autocommits_when_autocommit_is_off` feature is True
        """
        # print("schema:", sql)

        if self.connection.features.autocommits_when_autocommit_is_off:
            for tr in self.connection.connection.transactions:
                if tr.active:
                    tr.commit()
            # TransactionContext automatically commit statement
            with TransactionContext(self.connection.connection.trans()) as tr:
                try:
                    cur = tr.cursor()
                    cur.execute(cmd, params)
                except Exception as e:
                    raise e
        else:
            super(DatabaseSchemaEditor, self).execute(cmd, params)



# -*- utf-8 -*-

from django.db import models
# from django.utils.encoding import python_2_unicode_compatible


class FieldsTest(models.Model):
    pub_date = models.DateTimeField()
    mod_date = models.DateTimeField()


class BigS(models.Model):
    s = models.SlugField(max_length=255)


class Foo(models.Model):
    a = models.CharField(max_length=10)
    d = models.DecimalField(max_digits=5, decimal_places=3)


class Bar(models.Model):
    b = models.CharField(max_length=10)
    a = models.ForeignKey(Foo, related_name='bars', on_delete=models.CASCADE)


# @python_2_unicode_compatible
class DTModel(models.Model):
    name = models.CharField(max_length=32)
    start_datetime = models.DateTimeField(null=True, blank=True)
    end_datetime = models.DateTimeField(null=True, blank=True)
    start_date = models.DateField(null=True, blank=True)
    end_date = models.DateField(null=True, blank=True)
    start_time = models.TimeField(null=True, blank=True)
    end_time = models.TimeField(null=True, blank=True)
    duration = models.DurationField(null=True, blank=True)

    def __str__(self):
        return 'DTModel({0})'.format(self.name)



from urllib.request import urlopen
from json import loads, dumps
from typing import List
from urllib.parse import quote_plus

from wikibase.compiler import Cmd

charset_map = dict()

# https://github.com/yuvipanda/python-wdqs/blob/master/wdqs/client.py


class WbDatabase:
    SHRT_MIN = 1
    SHRT_MAX = 1
    INT_MIN = 1
    INT_MAX = 1
    LONG_MIN = 1
    LONG_MAX = 1

    ISOLATION_LEVEL_READ_COMMITED = 1

    _MEDIAWIKI_VERSION = '_mediawiki_version'
    _INSTANCE_OF = 'instance of'
    _SUBCLASS_OF = 'subclass of'
    _DJANGO_MODEL = 'django model'
    _SPARQL_ENDPOINT = '_sparql_endpoint'

    class Error:
        ...

    class DatabaseError(BaseException):
        ...

    class IntegrityError(BaseException):
        ...

    class OperationalError(BaseException):
        ...

    class DataError(BaseException):
        ...

    class InternalError(BaseException):
        ...

    class NotSupportedError(BaseException):
        ...

    class InterfaceError(BaseException):
        ...

    class ProgrammingError(BaseException):
        ...

    @staticmethod
    def connect(charset: str = 'utf8', url: str = '',
                user: str = '', password: str = '',
                instance_of_property_id: int = None,
                subclass_of_property_id: int = None,
                django_model_item_id: int = None,
                wdqs_sparql_endpoint: str = None):
        return WbDatabaseConnection(charset, url,
                                    user, password,
                                    instance_of_property_id,
                                    subclass_of_property_id,
                                    django_model_item_id,
                                    wdqs_sparql_endpoint)


class WbCursor:

    def close(self):
        print('close')

    def execute(self, cmd, params):
        if not isinstance(cmd, dict):
            raise WbDatabase.InternalError(f'Wrong command: {cmd}')
        print(f'execute ${cmd} with ${params}')

    def fetchall(self):
        """Fetch rows from the wikibase

        Returns:
            List[Tuple]: List of rows from the wikibase
        """
        return iter(())

    def fetchone(self):
        """Fetch single row from the wikibase

        Returns:
            Tuple: the row from the wikibase
        """
        return ()


class WbApi():

    def __init__(self, url: str, charset: str = 'utf-8'):
        super().__init__()
        self.url = url
        self.charset = charset

    def mediawiki_info(self):
        return loads(urlopen(
            f'{self.url}/api.php?action=query&meta=siteinfo&format=json').read().decode(self.charset))

    def search_entities(self, query):
        if 'label' in query:
            title_search_string = query['label']
            search_result = loads(urlopen(
                f'{self.url}/api.php?action=wbsearchentities&search={quote_plus(title_search_string)}&language=en&format=json').read().decode(self.charset))
            return search_result['search']
        raise WbDatabase.InternalError('Only search by label implemented')

    def new_item(self, data):
        search_result = loads(urlopen(
            f'{self.url}/api.php?action=wbeditentity&new=item&data={quote_plus(dumps(data))}&format=json').read().decode(self.charset))
        return search_result['search']

class WbDatabaseConnection:

    def __init__(self, charset: str, url: str,
                 user: str, password: str,
                 instance_of_property_id: int,
                 subclass_of_property_id: int,
                 django_model_item_id: int,
                 wdqs_sparql_endpoint: str):
        self.charset = charset
        self.api = WbApi(url, charset)
        self.user = user
        self.password = password  # TODO: hash instead plain
        if not django_model_item_id:
            # algo: 1. select django model by name 'django model'
            #       2. if not found create the item
            #       3. if found just get the identity
            django_model_item_id = self.create_entity_if_not_found_by_name_and_get_id_without_prefix(
                'django model')

        mediawiki_info = self.api.mediawiki_info()
        wikibase_conceptbaseuri = mediawiki_info['query']['general']['wikibase-conceptbaseuri']
        # https://avangard.testo.click/api.php?action=query&meta=siteinfo&siprop=extensions&format=json
        self.wikibase_info = {
            WbDatabase._MEDIAWIKI_VERSION: mediawiki_info['query']['general']['generator'],
            WbDatabase._INSTANCE_OF: f'{wikibase_conceptbaseuri}P{instance_of_property_id}',
            WbDatabase._SUBCLASS_OF: f'{wikibase_conceptbaseuri}P{subclass_of_property_id}',
            WbDatabase._DJANGO_MODEL: f'{wikibase_conceptbaseuri}Q{django_model_item_id}',
            WbDatabase._SPARQL_ENDPOINT: wdqs_sparql_endpoint if wdqs_sparql_endpoint else f'{url}/sparql'
        }
        self.transactions = []

    def create_entity_if_not_found_by_name_and_get_id_without_prefix(self, entity_name: str) -> int:
        entities = self.api.search_entities({'label': entity_name})
        if len(entities) == 1:
            return int(entities[0]['id'][1:])
        if len(entities) == 0:
            entity = self.api.new_entity({'title': entity_name})
            return int(entity['id'][1:])
        raise WbDatabase.InternalError(
            f'The entity {entity_name} has another one (i.e. not unique)')

    def db_info(self, key):
        return self.wikibase_info.get(key)

    def cursor(self):
        return WbCursor()

    def rollback(self):
        ...

    def commit(self):
        ...

    def close(self):
        ...

    def trans(self) -> List:
        return self


class TransactionContext:

    def __init__(self, connection: WbDatabaseConnection):
        self.connection = connection

    def __enter__(self):
        # make a database connection and return it
        ...
        return self.connection

    def __exit__(self, exc_type, exc_val, exc_tb):
        # make sure the dbconnection gets closed
        self.connection.close()
        ...



from unittest import TestCase

class UtilityTest(TestCase):

    def test_iter(self):
        result = [({'test': 1, 'd': 0}, ())]
        for a, b in result:
            self.assertTrue(True)
            print(f'a: {a}, b: {b}')
        self.assertTrue(True)



# Django settings for test_main project.
from os import environ

DEBUG = True
TEMPLATE_DEBUG = DEBUG


DATABASES = {
    'default': {
        'ENGINE': 'wikibase',
        'USER': 'SYSDBA',
        'PASSWORD': 'MasterKey',
        'URL': environ.get('WIKIBASE_URL', 'https://avangard.testo.click'),
        'OPTIONS': {
            'charset': 'utf-8',
            'instance_of_property_id': 1,
            'subclass_of_property_id': 2,
            'wdqs_sparql_endpoint': environ.get('SPARQL_ENDPOINT', 'https://avangard.testo.click/sparql')
        },

        'TEST': {
            'CHARSET': 'utf-8',
            'SERIALIZE': False,
            'PAGE_SIZE': 8192
        }
    },
    'other': {
        'ENGINE': 'wikibase',
        'USER': 'SYSDBA',
        'PASSWORD': 'MasterKey',
        'URL': environ.get('WIKIBASE_URL', 'https://avangard.testo.click'),
        'OPTIONS': {
            'charset': 'utf-8',
            'instance_of_property_id': 1,
            'subclass_of_property_id': 2,
            'wdqs_sparql_endpoint': environ.get('SPARQL_ENDPOINT', 'https://avangard.testo.click/sparql')
        },

        'TEST': {
            'CHARSET': 'utf-8',
            'SERIALIZE': False,
            'PAGE_SIZE': 8192
        }
    }

}

# Local time zone for this installation. Choices can be found here:
# http://en.wikipedia.org/wiki/List_of_tz_zones_by_name
# although not all choices may be available on all operating systems.
# In a Windows environment this must be set to your system time zone.
TIME_ZONE = 'America/Chicago'

# Language code for this installation. All choices can be found here:
# http://www.i18nguy.com/unicode/language-identifiers.html
LANGUAGE_CODE = 'en-us'

# If you set this to False, Django will make some optimizations so as not
# to load the internationalization machinery.
USE_I18N = True

# If you set this to False, Django will not format dates, numbers and
# calendars according to the current locale.
USE_L10N = True

# If you set this to False, Django will not use timezone-aware datetimes.
USE_TZ = False

# Absolute filesystem path to the directory that will hold user-uploaded files.
# Example: "/var/www/example.com/media/"
MEDIA_ROOT = ''

# URL that handles the media served from MEDIA_ROOT. Make sure to use a
# trailing slash.
# Examples: "http://example.com/media/", "http://media.example.com/"
MEDIA_URL = ''

# Absolute path to the directory static files should be collected to.
# Don't put anything in this directory yourself; store your static files
# in apps' "static/" subdirectories and in STATICFILES_DIRS.
# Example: "/var/www/example.com/static/"
STATIC_ROOT = ''

# URL prefix for static files.
# Example: "http://example.com/static/", "http://static.example.com/"
STATIC_URL = '/static/'

# Additional locations of static files
STATICFILES_DIRS = (
    # Put strings here, like "/home/html/static" or "C:/www/django/static".
    # Always use forward slashes, even on Windows.
    # Don't forget to use absolute paths, not relative paths.
)

# List of finder classes that know how to find static files in
# various locations.
STATICFILES_FINDERS = (
    'django.contrib.staticfiles.finders.FileSystemFinder',
    'django.contrib.staticfiles.finders.AppDirectoriesFinder',
    #    'django.contrib.staticfiles.finders.DefaultStorageFinder',
)

# Make this unique, and don't share it with anybody.
SECRET_KEY = ')!cccf)%m($bijkilb=z7-gjy_1!gj=v5^86(16%bl9fnr%ol0'

# List of callables that know how to import templates from various sources.
TEMPLATE_LOADERS = (
    'django.template.loaders.filesystem.Loader',
    'django.template.loaders.app_directories.Loader',
    #     'django.template.loaders.eggs.Loader',
)

MIDDLEWARE_CLASSES = (
    'django.middleware.common.CommonMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    # Uncomment the next line for simple clickjacking protection:
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
)

ROOT_URLCONF = 'test_main.urls'

# Python dotted path to the WSGI application used by Django's runserver.
WSGI_APPLICATION = 'test_main.wsgi.application'

TEMPLATE_DIRS = (
    # Put strings here, like "/home/html/django_templates" or "C:/www/django/templates".
    # Always use forward slashes, even on Windows.
    # Don't forget to use absolute paths, not relative paths.
)

INSTALLED_APPS = (
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.sites',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    # Uncomment the next line to enable the admin:
    # 'django.contrib.admin',
    # Uncomment the next line to enable admin documentation:
    # 'django.contrib.admindocs',
    'tests.test_main.test_base',
    'tests.test_main.model_fields',
    'tests.test_main.many_to_one',
    'tests.test_main.transactions',
    'tests.test_main.schema',
    'tests.test_main.introspection',
    'tests.test_main.inspectdb',
    'tests.test_main.datatypes',
    'tests.test_main.dates',
    'tests.test_main.datetimes',
    'tests.test_main.lookup',
    'tests.test_main.expressions',
)

# A sample logging configuration. The only tangible logging
# performed by this configuration is to send an email to
# the site admins on every HTTP 500 error when DEBUG=False.
# See http://docs.djangoproject.com/en/dev/topics/logging for
# more details on how to customize your logging configuration.
LOGGING = {
    'version': 1,
    'disable_existing_loggers': False,
    'filters': {
        'require_debug_false': {
            '()': 'django.utils.log.RequireDebugFalse'
        }
    },
    'handlers': {
        'mail_admins': {
            'level': 'ERROR',
            'filters': ['require_debug_false'],
            'class': 'django.utils.log.AdminEmailHandler'
        }
    },
    'loggers': {
        'django.request': {
            'handlers': ['mail_admins'],
            'level': 'ERROR',
            'propagate': True,
        },
    }
}



from typing import Any
from django.db.models.sql import compiler
from json.encoder import JSONEncoder

# The izip_longest was renamed to zip_longest in py3
try:
    from itertools import zip_longest
except ImportError:
    from itertools import izip_longest as zip_longest

from logging import debug
from json import dumps


class Cmd(dict):

    def __init__(self, cmd: str, data: dict = None):
        dict.__init__(self, cmd=cmd, data=data)

    def __repr__(self) -> str:
        return dumps(self)


class SQLCompiler(compiler.SQLCompiler):

    _select_data = {
        'sparql': None
    }

    def resolve_columns(self, row, fields=()):
        # We need to convert values from database to correct django field representation.
        # For instance, if we defined a BooleanField field, django-wikibase do create a
        # smallint field into DB. When retrieving this field value, it's converted to
        # BooleanField again.
        index_start = len(self.query.extra_select)
        values = []
        for value, field in zip_longest(row[index_start:], fields):
            v = self.query.convert_values(
                value, field, connection=self.connection)
            values.append(v)
        return row[:index_start] + tuple(values)

    def as_sql(self, with_limits=True, with_col_aliases=False):
        sql, params = super(SQLCompiler, self).as_sql(
            with_limits=False, with_col_aliases=with_col_aliases)
        debug('SQL compiler, as_sql, with limits {}, with col alizses {}',
              with_limits, with_col_aliases)

        if with_limits:
            limits = []
            if self.query.high_mark is not None:
                limits.append('FIRST %d' %
                              (self.query.high_mark - self.query.low_mark))
            if self.query.low_mark:
                if self.query.high_mark is None:
                    val = self.connection.ops.no_limit_value()
                    if val:
                        limits.append('FIRST %d' % val)
                limits.append('SKIP %d' % self.query.low_mark)
            sql = 'SELECT %s %s' % (' '.join(limits), sql[6:].strip())
        cmd_data = {**self._select_data, **{'sparql': sql}}
        return Cmd('select', cmd_data), params


class SQLInsertCompiler(compiler.SQLInsertCompiler, SQLCompiler):

    _add_items_data = {
        'sparql': None
    }

    def as_sql(self, with_limits=True, with_col_aliases=False):
        sql, params = super(compiler.SQLInsertCompiler, self).as_sql(
            with_limits=False, with_col_aliases=with_col_aliases)
        debug('SQL insert compiler, as_sql, with limits {}, with col alizses {}',
              with_limits, with_col_aliases)
        cmd_data = {**self._add_items_data, **{'sparql': sql}}
        return [(Cmd('add_items', cmd_data), (params))]


class SQLDeleteCompiler(compiler.SQLDeleteCompiler, SQLCompiler):

    _remove_items_data = {
        'sparql': None
    }

    def as_sql(self, with_limits=True, with_col_aliases=False):
        sql, params = super(compiler.SQLDeleteCompiler, self).as_sql(
            with_limits=False, with_col_aliases=with_col_aliases)
        debug('SQL delete compiler, as_sql, with limits {}, with col alizses {}',
              with_limits, with_col_aliases)
        cmd_data = {**self._remove_items_data, **{'sparql': sql}}
        return [(Cmd('remove_items', cmd_data), (params))]


class SQLUpdateCompiler(compiler.SQLUpdateCompiler, SQLCompiler):

    _set_items_data = {
        'sparql': None
    }

    def as_sql(self, with_limits=True, with_col_aliases=False):
        sql, params = super(compiler.SQLUpdateCompiler, self).as_sql(
            with_limits=False, with_col_aliases=with_col_aliases)
        debug('SQL update compiler, as_sql, with limits {}, with col alizses {}',
              with_limits, with_col_aliases)
        cmd_data = {**self._set_items_data, **{'sparql': sql}}
        return [(Cmd('set_items', cmd_data), (params))]


class SQLAggregateCompiler(compiler.SQLAggregateCompiler, SQLCompiler):

    _agg_items_data = {
        'sparql': None
    }

    def as_sql(self, with_limits=True, with_col_aliases=False):
        sql, params = super(compiler.SQLAggregateCompiler, self).as_sql(
            with_limits=False, with_col_aliases=with_col_aliases)
        debug('SQL aggregate compiler, as_sql, with limits {}, with col alizses {}',
              with_limits, with_col_aliases)
        cmd_data = {**self._agg_items_data, **{'sparql': sql}}
        return [(Cmd('agg_items', cmd_data), (params))]




__API_REGISTRY = dict()


def api(api_identity: str):
    api_module = __API_REGISTRY[api_identity] if api_identity in __API_REGISTRY else None
    if api_module:
        return api_module
    api_module = Api(api_identity)
    __API_REGISTRY[api_identity] = api_module
    return api_module


class Api:

    def __init__(self, id: str):
        super().__init__()
        self.url = 'http://localhost:8081'
        self.id = id



# -*- coding: utf-8 -*-

from datetime import datetime, timedelta

from django.conf import settings
from django.db import connection, DatabaseError
from django.db.models import F, DateField, DateTimeField, IntegerField, TimeField, CASCADE
from django.db.models.fields.related import ForeignKey
from django.db.models.functions import (
    Extract, ExtractDay, ExtractHour, ExtractMinute, ExtractMonth,
    ExtractSecond, ExtractWeek, ExtractWeekDay, ExtractYear, Trunc, TruncDate,
    TruncDay, TruncHour, TruncMinute, TruncMonth, TruncSecond, TruncTime,
    TruncYear,
)
from django.test import TestCase, TransactionTestCase, override_settings
from django.utils import timezone


from tests.test_main.test_base.models import BigS, FieldsTest, Foo, Bar, DTModel


def microsecond_support(value):
    return value if connection.features.supports_microsecond_precision else value.replace(microsecond=0)


def truncate_to(value, kind, tzinfo=None):
    # Convert to target timezone before truncation
    if tzinfo is not None:
        value = value.astimezone(tzinfo)

    def truncate(value, kind):
        if kind == 'second':
            return value.replace(microsecond=0)
        if kind == 'minute':
            return value.replace(second=0, microsecond=0)
        if kind == 'hour':
            return value.replace(minute=0, second=0, microsecond=0)
        if kind == 'day':
            if isinstance(value, datetime):
                return value.replace(hour=0, minute=0, second=0, microsecond=0)
            return value
        if kind == 'month':
            if isinstance(value, datetime):
                return value.replace(day=1, hour=0, minute=0, second=0, microsecond=0)
            return value.replace(day=1)
        # otherwise, truncate to year
        if isinstance(value, datetime):
            return value.replace(month=1, day=1, hour=0, minute=0, second=0, microsecond=0)
        return value.replace(month=1, day=1)

    value = truncate(value, kind)
    if tzinfo is not None:
        # If there was a daylight saving transition, then reset the timezone.
        value = timezone.make_aware(value.replace(tzinfo=None), tzinfo)
    return value


class WikibaseTest(TestCase):
    def setUp(self):
        pass

    def test_server_version(self):
        version = connection.server_version
        self.assertNotEqual(version, '')

    def test_mediawiki_version(self):
        version = connection.ops.mediawiki_version
        self.assertNotEqual(version, [])


class DatabaseOperationsTest(TestCase):
    def setUp(self):
        self.ops = connection.ops

    def test_get_sequence_name(self):
        sq_name = self.ops.get_sequence_name('TEST')
        self.assertEqual(sq_name, '"TEST_SQ"')

    def test_drop_sequence_sql(self):
        sql = self.ops.drop_sequence_sql('TEST')
        self.assertEqual(sql, 'DROP SEQUENCE "TEST_SQ"')

    def test_date_extract_sql(self):
        sql = self.ops.date_extract_sql('week_day', 'DATE_FIELD')
        value = "EXTRACT(WEEKDAY FROM DATE_FIELD) + 1"
        self.assertEqual(sql, value)

        sql = self.ops.date_extract_sql('year', 'DATE_FIELD')
        value = "EXTRACT(YEAR FROM DATE_FIELD)"
        self.assertEqual(sql, value)

        sql = self.ops.date_extract_sql('month', 'DATE_FIELD')
        value = "EXTRACT(MONTH FROM DATE_FIELD)"
        self.assertEqual(sql, value)

        sql = self.ops.date_extract_sql('day', 'DATE_FIELD')
        value = "EXTRACT(DAY FROM DATE_FIELD)"
        self.assertEqual(sql, value)

    def test_datetime_trunc_sql(self):
        sql = self.ops.datetime_trunc_sql('year', 'DATE_FIELD', None)
        value = "CAST(EXTRACT(year FROM DATE_FIELD)||'-01-01 00:00:00' AS TIMESTAMP)"
        self.assertEqual(sql, value)

        sql = self.ops.datetime_trunc_sql('month', 'DATE_FIELD', None)
        value = "CAST(EXTRACT(year FROM DATE_FIELD)||'-'||EXTRACT(month FROM DATE_FIELD)||'-01 00:00:00' AS TIMESTAMP)"
        self.assertEqual(sql, value)

        sql = self.ops.datetime_trunc_sql('day', 'DATE_FIELD', None)
        value = "CAST(EXTRACT(year FROM DATE_FIELD)||'-'||EXTRACT(month FROM DATE_FIELD)||'-'||EXTRACT(day FROM DATE_FIELD)||' 00:00:00' AS TIMESTAMP)"
        self.assertEqual(sql, value)

        sql = self.ops.datetime_trunc_sql('hour', 'DATE_FIELD', None)
        value = "CAST(EXTRACT(year FROM DATE_FIELD)||'-'||EXTRACT(month FROM DATE_FIELD)||'-'||EXTRACT(day FROM DATE_FIELD)||' '||EXTRACT(hour FROM DATE_FIELD)||':00:00' AS TIMESTAMP)"
        self.assertEqual(sql, value)

        sql = self.ops.datetime_trunc_sql('minute', 'DATE_FIELD', None)
        value = "CAST(EXTRACT(year FROM DATE_FIELD)||'-'||EXTRACT(month FROM DATE_FIELD)||'-'||EXTRACT(day FROM DATE_FIELD)||' '||EXTRACT(hour FROM DATE_FIELD)||':'||EXTRACT(minute FROM DATE_FIELD)||':00' AS TIMESTAMP)"
        self.assertEqual(sql, value)

        sql = self.ops.datetime_trunc_sql('second', 'DATE_FIELD', None)
        value = "CAST(EXTRACT(year FROM DATE_FIELD)||'-'||EXTRACT(month FROM DATE_FIELD)||'-'||EXTRACT(day FROM DATE_FIELD)||' '||EXTRACT(hour FROM DATE_FIELD)||':'||EXTRACT(minute FROM DATE_FIELD)||':'||TRUNC(EXTRACT(second FROM DATE_FIELD)) AS TIMESTAMP)"
        self.assertEqual(sql, value)

    def test_time_trunc_sql(self):
        sql = self.ops.time_trunc_sql('hour', 'TIME_FIELD')
        out = "CAST(EXTRACT(hour FROM TIME_FIELD) || ':00:00' AS TIME)"
        self.assertEqual(sql, out)

        sql = self.ops.time_trunc_sql('minute', 'TIME_FIELD')
        out = "CAST(EXTRACT(hour FROM TIME_FIELD) || ':' || EXTRACT(minute FROM TIME_FIELD) || ':00' AS TIME)"
        self.assertEqual(sql, out)

        sql = self.ops.time_trunc_sql('second', 'TIME_FIELD')
        out = "CAST(EXTRACT(hour FROM TIME_FIELD) || ':' || EXTRACT(minute FROM TIME_FIELD) || ':' || TRUNC(EXTRACT(second FROM TIME_FIELD)) AS TIME)"
        self.assertEqual(sql, out)


class DatabaseSchemaTests(TransactionTestCase):
    def test_no_index_for_foreignkey(self):
        """
        WikibaseSQL already creates indexes automatically for foreign keys. (#70).
        """
        index_sql = connection.schema_editor()._model_indexes_sql(Bar)
        self.assertEqual(index_sql, [])

    def test_fk_index_creation(self):
        new_field = ForeignKey(Foo, on_delete=CASCADE)
        new_field.set_attributes_from_name(None)
        with connection.schema_editor() as editor:
            editor.add_field(
                Bar,
                new_field
            )
            # Just return indexes others that not automaically created by Fk
            indexes = editor._get_field_indexes(Bar, new_field)
        self.assertEqual(indexes, [])

    def test_fk_remove_issue70(self):
        with connection.schema_editor() as editor:
            editor.remove_field(
                Bar,
                Bar._meta.get_field("a")
            )
        self.assertRaises(DatabaseError)


class SlugFieldTests(TestCase):
    def test_slugfield_max_length(self):
        """
        Make sure SlugField honors max_length (#9706)
        """
        bs = BigS.objects.create(s='slug' * 50)
        bs = BigS.objects.get(pk=bs.pk)
        self.assertEqual(bs.s, 'slug' * 50)


class DateFieldTests(TestCase):
    def tests_date_interval(self):
        obj = FieldsTest()
        obj.pub_date = datetime.now()
        obj.mod_date = obj.pub_date + timedelta(days=3)
        obj.save()

        objs = FieldsTest.objects.filter(mod_date__gte=F('pub_date') + timedelta(days=3)).all()
        self.assertEqual(len(objs), 1)


@override_settings(USE_TZ=False)
class DateFunctionTests(TestCase):

    def create_model(self, start_datetime, end_datetime):
        return DTModel.objects.create(
            name=start_datetime.isoformat(),
            start_datetime=start_datetime, end_datetime=end_datetime,
            start_date=start_datetime.date(), end_date=end_datetime.date(),
            start_time=start_datetime.time(), end_time=end_datetime.time(),
            duration=(end_datetime - start_datetime),
        )

    def test_trunc_func(self):
        start_datetime = microsecond_support(datetime(2015, 6, 15, 14, 30, 50, 321))
        end_datetime = microsecond_support(datetime(2016, 6, 15, 14, 10, 50, 123))
        if settings.USE_TZ:
            start_datetime = timezone.make_aware(start_datetime, is_dst=False)
            end_datetime = timezone.make_aware(end_datetime, is_dst=False)
        self.create_model(start_datetime, end_datetime)
        self.create_model(end_datetime, start_datetime)

        msg = 'output_field must be either DateField, TimeField, or DateTimeField'
        with self.assertRaisesMessage(ValueError, msg):
            list(DTModel.objects.annotate(truncated=Trunc('start_datetime', 'year', output_field=IntegerField())))

        with self.assertRaisesMessage(AssertionError, "'name' isn't a DateField, TimeField, or DateTimeField."):
            list(DTModel.objects.annotate(truncated=Trunc('name', 'year', output_field=DateTimeField())))

        with self.assertRaisesMessage(ValueError, "Cannot truncate DateField 'start_date' to DateTimeField"):
            list(DTModel.objects.annotate(truncated=Trunc('start_date', 'second')))

        with self.assertRaisesMessage(ValueError, "Cannot truncate TimeField 'start_time' to DateTimeField"):
            list(DTModel.objects.annotate(truncated=Trunc('start_time', 'month')))

        with self.assertRaisesMessage(ValueError, "Cannot truncate DateField 'start_date' to DateTimeField"):
            list(DTModel.objects.annotate(truncated=Trunc('start_date', 'month', output_field=DateTimeField())))

        with self.assertRaisesMessage(ValueError, "Cannot truncate TimeField 'start_time' to DateTimeField"):
            list(DTModel.objects.annotate(truncated=Trunc('start_time', 'second', output_field=DateTimeField())))

        def test_datetime_kind(kind):
            self.assertQuerysetEqual(
                DTModel.objects.annotate(
                    truncated=Trunc('start_datetime', kind, output_field=DateTimeField())
                ).order_by('start_datetime'),
                [
                    (truncate_to(start_datetime, kind)),
                    (truncate_to(end_datetime, kind))
                ],
                lambda m: (m.truncated)
            )

        def test_date_kind(kind):
            self.assertQuerysetEqual(
                DTModel.objects.annotate(
                    truncated=Trunc('start_date', kind, output_field=DateField())
                ).order_by('start_datetime'),
                [
                    (truncate_to(start_datetime.date(), kind)),
                    (truncate_to(end_datetime.date(), kind))
                ],
                lambda m: (m.truncated)
            )

        def test_time_kind(kind):
            self.assertQuerysetEqual(
                DTModel.objects.annotate(
                    truncated=Trunc('start_time', kind, output_field=TimeField())
                ).order_by('start_datetime'),
                [
                    (truncate_to(start_datetime.time(), kind)),
                    (truncate_to(end_datetime.time(), kind))
                ],
                lambda m: (m.truncated)
            )

        test_date_kind('year')
        test_date_kind('month')
        test_date_kind('day')
        test_time_kind('hour')
        test_time_kind('minute')
        test_time_kind('second')
        test_datetime_kind('year')
        test_datetime_kind('month')
        test_datetime_kind('day')
        test_datetime_kind('hour')
        test_datetime_kind('minute')
        test_datetime_kind('second')

        qs = DTModel.objects.filter(start_datetime__date=Trunc('start_datetime', 'day', output_field=DateField()))
        self.assertEqual(qs.count(), 2)

    def test_trunc_time_func(self):
        start_datetime = microsecond_support(datetime(2015, 6, 15, 14, 30, 50, 321000))
        end_datetime = microsecond_support(datetime(2016, 6, 15, 14, 10, 50, 123000))
        if settings.USE_TZ:
            start_datetime = timezone.make_aware(start_datetime, is_dst=False)
            end_datetime = timezone.make_aware(end_datetime, is_dst=False)
        self.create_model(start_datetime, end_datetime)
        self.create_model(end_datetime, start_datetime)
        self.assertQuerysetEqual(
            DTModel.objects.annotate(extracted=TruncTime('start_datetime')).order_by('start_datetime'),
            [
                (start_datetime.time()),
                (end_datetime.time()),
            ],
            lambda m: (m.extracted)
        )
        self.assertEqual(DTModel.objects.filter(start_datetime__time=TruncTime('start_datetime')).count(), 2)

        with self.assertRaisesMessage(ValueError, "Cannot truncate DateField 'start_date' to TimeField"):
            list(DTModel.objects.annotate(truncated=TruncTime('start_date')))

        with self.assertRaisesMessage(ValueError, "Cannot truncate DateField 'start_date' to TimeField"):
            list(DTModel.objects.annotate(truncated=TruncTime('start_date', output_field=DateField())))



"""
Wikibase database backend for Django.
"""

import sys

from wikibase.wdb import WbDatabase
from wikibase.wdb import charset_map

from django.db import utils
from django.db.backends.base.base import BaseDatabaseWrapper

from django.utils.encoding import smart_str
from django.utils.functional import cached_property
from six import reraise

from .operations import DatabaseOperations
from .features import DatabaseFeatures
from .client import DatabaseClient
from .creation import DatabaseCreation
from .introspection import DatabaseIntrospection
from .schema import DatabaseSchemaEditor
from .validation import DatabaseValidation


DatabaseError = WbDatabase.DatabaseError
IntegrityError = WbDatabase.IntegrityError
OperationalError = WbDatabase.OperationalError


class DatabaseWrapper(BaseDatabaseWrapper):
    vendor = 'mast.eu.spb.ru'

    # This dictionary maps Field objects to their associated Wikibase column
    # types, as strings. Column-type strings can contain format strings; they'll
    # be interpolated against the values of Field.__dict__ before being output.
    # If a column type is set to None, it won't be included in the output.
    #
    # Any format strings starting with "qn_" are quoted before being used in the
    # output (the "qn_" prefix is stripped before the lookup is performed.

    data_types = {
        'AutoField': 'integer',
        'BigAutoField': 'bigint',
        'BinaryField': 'blob sub_type 0',
        'BooleanField': 'smallint',  # for wikibase 3 it changes in init_connection_state
        'CharField': 'varchar(%(max_length)s)',
        'CommaSeparatedIntegerField': 'varchar(%(max_length)s)',
        'DateField': 'date',
        'DateTimeField': 'timestamp',
        'DecimalField': 'decimal(%(max_digits)s, %(decimal_places)s)',
        'DurationField': 'bigint',
        'FileField': 'varchar(%(max_length)s)',
        'FilePathField': 'varchar(%(max_length)s)',
        'FloatField': 'double precision',
        'IntegerField': 'integer',
        'BigIntegerField': 'bigint',
        'IPAddressField': 'char(15)',
        'GenericIPAddressField': 'char(39)',
        # for wikibase 3 it changes in init_connection_state
        'NullBooleanField': 'smallint',
        'OneToOneField': 'integer',
        'PositiveIntegerField': 'integer',
        'PositiveSmallIntegerField': 'smallint',
        'SlugField': 'varchar(%(max_length)s)',
        'SmallIntegerField': 'smallint',
        'TextField': 'blob sub_type 1',
        'TimeField': 'time',
        'UUIDField': 'char(32)',
    }

    data_type_check_constraints = {
        # for wikibase 3 it changes in init_connection_state
        'BooleanField': '%(qn_column)s IN (0,1)',
        'NullBooleanField': '(%(qn_column)s IN (0,1)) OR (%(qn_column)s IS NULL)',
        'PositiveIntegerField': '%(qn_column)s >= 0',
        'PositiveSmallIntegerField': '%(qn_column)s >= 0',
    }

    operators = {
        'exact': '= %s',
        'iexact': '= UPPER(%s)',
        'contains': "LIKE %s ESCAPE'\\'",
        # 'CONTAINING %s', #case is ignored
        'icontains': "LIKE UPPER(%s) ESCAPE'\\'",
        'gt': '> %s',
        'gte': '>= %s',
        'lt': '< %s',
        'lte': '<= %s',
        # 'STARTING WITH %s', #looks to be faster than LIKE
        'startswith': "LIKE %s ESCAPE'\\'",
        'endswith': "LIKE %s ESCAPE'\\'",
        # 'STARTING WITH UPPER(%s)',
        'istartswith': "LIKE UPPER(%s) ESCAPE'\\'",
        'iendswith': "LIKE UPPER(%s) ESCAPE'\\'",
        'regex': "SIMILAR TO %s",
        'iregex': "SIMILAR TO %s",  # Case Sensitive depends on collation
    }

    # The patterns below are used to generate SQL pattern lookup clauses when
    # the right-hand side of the lookup isn't a raw string (it might be an expression
    # or the result of a bilateral transformation).
    # In those cases, special characters for LIKE operators (e.g. \, *, _) should be
    # escaped on database side.
    #
    # Note: we use str.format() here for readability as '%' is used as a wildcard for
    # the LIKE operator.
    pattern_esc = r"REPLACE(REPLACE(REPLACE({}, '\', '\\'), '%%', '\%%'), '_', '\_')"
    pattern_ops = {
        'contains': "LIKE '%%' || {} || '%%'",
        'icontains': "LIKE '%%' || UPPER({}) || '%%'",
        'startswith': "LIKE {} || '%%'",
        'istartswith': "LIKE UPPER({}) || '%%'",
        'endswith': "LIKE '%%' || {}",
        'iendswith': "LIKE '%%' || UPPER({})",
    }

    Database = WbDatabase
    SchemaEditorClass = DatabaseSchemaEditor

    # Classes instantiated in __init__().
    client_class = DatabaseClient
    creation_class = DatabaseCreation
    features_class = DatabaseFeatures
    introspection_class = DatabaseIntrospection
    ops_class = DatabaseOperations

    def __init__(self, *args, **kwargs):
        super(DatabaseWrapper, self).__init__(*args, **kwargs)

        self._server_version = None
        self._db_charset = None
        self.encoding = None

        opts = self.settings_dict["OPTIONS"]
        RC = WbDatabase.ISOLATION_LEVEL_READ_COMMITED
        self.isolation_level = opts.get('isolation_level', RC)

        self.features = DatabaseFeatures(self)
        self.ops = DatabaseOperations(self)
        self.client = DatabaseClient(self)
        self.creation = DatabaseCreation(self)
        self.introspection = DatabaseIntrospection(self)
        self.validation = DatabaseValidation(self)

    # #### Backend-specific methods for creating connections and cursors #####

    def get_connection_params(self):
        """Returns a dict of parameters suitable for get_new_connection."""
        settings_dict = self.settings_dict
        if settings_dict['URL'] == '':
            from django.core.exceptions import ImproperlyConfigured
            raise ImproperlyConfigured(
                "settings.DATABASES is improperly configured. "
                "Please supply the URL value (for example https://www.wikidata.org/w). "
                "URL value is a prefix for the api.php")

        # The port param is not used by fdb. It must be setting by url string
        url = settings_dict['URL']
        conn_params = {
            'charset': 'utf-8',
        }
        conn_params['url'] = url % settings_dict
        if settings_dict['USER']:
            conn_params['user'] = settings_dict['USER']
        if settings_dict['PASSWORD']:
            conn_params['password'] = settings_dict['PASSWORD']
        if 'ROLE' in settings_dict:
            conn_params['role'] = settings_dict['ROLE']
        options = settings_dict['OPTIONS'].copy()
        conn_params.update(options)

        self._db_charset = conn_params.get('charset')
        self.encoding = charset_map.get(self._db_charset, 'utf_8')

        return conn_params

    def get_new_connection(self, conn_params):
        """Opens a connection to the database."""
        return WbDatabase.connect(**conn_params)

    def init_connection_state(self):
        """Initializes the database connection settings."""
        # if int(self.ops.mediawiki_version[3]) >= 3:
        #     self.data_types['BooleanField'] = 'boolean'
        #     self.data_types['NullBooleanField'] = 'boolean'
        #     self.data_type_check_constraints['BooleanField'] = '%(qn_column)s IN (False,True)'
        #     self.data_type_check_constraints['NullBooleanField'] = '(%(qn_column)s IN (False,True)) OR (%(qn_column)s IS NULL)'
        ...

    def create_cursor(self, name=None):
        """Creates a cursor. Assumes that a connection is established."""
        cursor = self.connection.cursor()
        return WikibaseCursorWrapper(cursor, self.encoding)

    # ##### Foreign key constraints checks handling #####

    def disable_constraint_checking(self):
        """
        Backends can implement as needed to temporarily disable foreign key
        constraint checking. Should return True if the constraints were
        disabled and will need to be reenabled.
        """
        self.disable_constraints()
        return True

    def enable_constraint_checking(self):
        """
        Backends can implement as needed to re-enable foreign key constraint
        checking.
        """
        self.enable_constraints()

    def disable_constraints(self):
        """
        Disables restrictions such as foreign keys, checks and unique.

        .. important::

           The server does not support explicit disabling of restrictions,
           therefore, implementation is made using additional tables.

        Note:
           If there is an error when disable restrictions an exception is thrown.
        """
        create_django_constraint = """
            create table django$constraint (
                django$constraint_name varchar(31) not null constraint pk_djangocopyconstraint primary key,
                django$constraint_type varchar(11) not null,
                django$relation_name varchar(31) not null,
                django$index_name varchar(31),
                django$const_name_uq varchar(31),
                django$update_rule varchar(11),
                django$delete_rule varchar(11),
                django$constraint_source blob sub_type 1
            )
        """
        create_django_constraint_segment = """
            create table django$constraint_segment (
                django$constraint_name varchar(31) not null,
                django$field_name varchar(31) not null,
                django$position integer not null,
                constraint pk_djangocopyconstraintsegment primary key (django$constraint_name, django$field_name)
            )
        """
        save_constraints = """
            merge into django$constraint dc
            using (select
                trim(trailing from c.rdb$constraint_name) as constraint_name,
                trim(trailing from c.rdb$constraint_type) as constraint_type,
                trim(trailing from c.rdb$relation_name) as relation_name,
                nullif(trim(trailing from c.rdb$index_name), '') as index_name,
                t.rdb$trigger_source as constraint_source,
                nullif(trim(trailing from r.rdb$const_name_uq), '') as const_name_uq,
                nullif(trim(trailing from r.rdb$update_rule), '') as update_rule,
                nullif(trim(trailing from r.rdb$delete_rule), '') as delete_rule
            from 
            rdb$relation_constraints c left join
                (select * from rdb$check_constraints p
                where p.rdb$trigger_name = (select first 1 rdb$trigger_name from rdb$check_constraints o
                where p.rdb$constraint_name = o.rdb$constraint_name)) h on c.rdb$constraint_name = h.rdb$constraint_name
                left join rdb$triggers t on t.rdb$trigger_name = h.rdb$trigger_name
                left join rdb$ref_constraints r on r.rdb$constraint_name = c.rdb$constraint_name
            where c.rdb$constraint_type in ('FOREIGN KEY', 'CHECK', 'UNIQUE')
            and
            c.rdb$relation_name not starting with 'RDB$') rc
            on dc.django$constraint_name = rc.constraint_name
            when matched then
                update set
                    dc.django$constraint_type = rc.constraint_type,
                    dc.django$relation_name = rc.relation_name,
                    dc.django$index_name = rc.index_name,
                    dc.django$constraint_source = rc.constraint_source,
                    dc.django$const_name_uq = rc.const_name_uq,
                    dc.django$update_rule = rc.update_rule,
                    dc.django$delete_rule = rc.delete_rule
            when not matched then
                insert (dc.django$constraint_name,
                    dc.django$constraint_type,
                    dc.django$relation_name,
                    dc.django$index_name,
                    dc.django$constraint_source,
                    dc.django$const_name_uq,
                    dc.django$update_rule,
                    dc.django$delete_rule)
                values (rc.constraint_name,
                    rc.constraint_type,
                    rc.relation_name,
                    rc.index_name,
                    rc.constraint_source,
                    rc.const_name_uq,
                    update_rule,
                    rc.delete_rule)
        """
        save_segment_constraints = """
            merge into django$constraint_segment dcs
            using (
                select c.rdb$constraint_name as constraint_name,
                    s.rdb$field_name as field_name,
                    s.rdb$field_position as field_position
                from rdb$relation_constraints c, rdb$index_segments s
                where s.rdb$index_name = c.rdb$index_name
                and c.rdb$constraint_type in ('FOREIGN KEY', 'CHECK', 'UNIQUE')
                and c.rdb$relation_name not starting with 'RDB$') cs
            on dcs.django$constraint_name = cs.constraint_name
                and dcs.django$field_name = cs.field_name
            when matched then
                update set
                    dcs.django$position = cs.field_position
            when not matched then
                insert (django$constraint_name,
                    django$field_name,
                    django$position)
                values (
                    cs.constraint_name,
                    cs.field_name,
                    cs.field_position)
        """
        select_drop_constraints = """
            select trim(trailing from rdb$constraint_name) as constr_name,
                trim(trailing from c.rdb$relation_name) as table_name
            from rdb$relation_constraints c
            where c.rdb$constraint_type in ('FOREIGN KEY', 'CHECK', 'UNIQUE')
            and
            c.rdb$relation_name not starting with 'RDB$'
            order by c.rdb$constraint_type
        """
        editor = self.schema_editor()
        if not self.table_exists("django$constraint"):
            editor.execute(create_django_constraint)
        if not self.table_exists("django$constraint_segment"):
            editor.execute(create_django_constraint_segment)
        editor.execute(save_constraints)
        editor.execute(save_segment_constraints)
        constraints = self.get_drop_constraints(select_drop_constraints)
        for hm in constraints:
            sql = """
                alter table "%(table)s" drop constraint "%(constraint)s"
            """ % {
                'table': hm['TABLE_NAME'],
                'constraint': hm['CONSTR_NAME']
            }
            editor.execute(sql)

    def enable_constraints(self):
        """
        Enables restrictions that have been disabled by calling the method `disable_constraint_checking`.

        .. important::

           The server does not support explicit disabling of restrictions,
           therefore, implementation is made using additional tables.

        Note:
           If there is an error when enable restrictions an exception is thrown.
        """
        editor = self.schema_editor()
        select_django_constraint = """
        select django$constraint_name as constraint_name,
            django$constraint_type as constraint_type,
            django$relation_name as relation_name,
            django$index_name as index_name,
            django$constraint_source as constraint_source,
            django$const_name_uq as const_name_uq,
            django$update_rule as update_rule,
            django$delete_rule as delete_rule
        from django$constraint
        order by django$constraint_type desc"""
        constraints = self.get_drop_constraints(select_django_constraint)
        for hm in constraints:
            if not self.table_exists(hm['RELATION_NAME'].strip()):
                continue
            create_string = "alter table \"" + hm['RELATION_NAME'] + "\" add "
            if not hm['CONSTRAINT_NAME'].startswith("RDB$"):
                create_string += "constraint " + hm['CONSTRAINT_NAME']
            if hm['CONSTRAINT_TYPE'].casefold() == "CHECK".casefold():
                create_string += " " + hm['CONSTRAINT_SOURCE']
            elif hm['CONSTRAINT_TYPE'].casefold() == "FOREIGN KEY".casefold():
                select_relation = """select coalesce(trim(trailing from rdb$relation_name), '')
                    from rdb$relation_constraints where rdb$constraint_name = '%s'                    
                """ % hm['CONST_NAME_UQ']
                table = ''
                with self.cursor() as cursor:
                    cursor.execute(select_relation)
                    res = cursor.fetchone()
                    if not res:
                        continue
                    table = '"' + res[0].strip() + '"'
                select_django_constraint_segment = """
                    select django$field_name as field_name from django$constraint_segment s 
                    where s.django$constraint_name = '%s'
                    order by django$position
                """ % hm['CONSTRAINT_NAME']
                segments = []
                with self.cursor() as cursor:
                    cursor.execute(select_django_constraint_segment)
                    for row in cursor.fetchall():
                        map = {}
                        for i, desc in enumerate(cursor.cursor.cursor.description):
                            map[desc[0]] = row[i]
                        segments.append(map)
                field_list = []
                for field in segments:
                    field_list.append('"' + field['FIELD_NAME'].strip() + '"')
                create_string += " foreign key(" + ','.join(str(x) for x in field_list) +\
                                 ") references " + table + "("
                select_index_segment = """
                    select trim(trailing from rdb$field_name) as field_name from rdb$index_segments s
                    join rdb$relation_constraints r on r.rdb$index_name = s.rdb$index_name
                    where r.rdb$constraint_name = '%s'
                    order by rdb$field_position
                """ % hm['CONST_NAME_UQ']
                index_segments = []
                with self.cursor() as cursor:
                    cursor.execute(select_index_segment)
                    for row in cursor.fetchall():
                        index_segments.append('"' + row[0].strip() + '"')
                create_string += ','.join(str(x) for x in index_segments) + ")"
                if hm['UPDATE_RULE'].casefold() != "RESTRICT".casefold():
                    create_string += " on update " + hm['UPDATE_RULE']
                if hm['DELETE_RULE'].casefold() != "RESTRICT".casefold():
                    create_string += " on delete " + hm['DELETE_RULE']
                if not hm['INDEX_NAME'].startswith("RDB$"):
                    create_string += " using index " + hm['INDEX_NAME']
            elif hm['CONSTRAINT_TYPE'].casefold() == "UNIQUE".casefold():
                select_django_constraint_segment = """
                    select django$field_name as field_name from django$constraint_segment s
                    where s.django$constraint_name = '%s'
                    order by django$position
                """ % hm['CONSTRAINT_NAME']
                segments = []
                with self.cursor() as cursor:
                    cursor.execute(select_django_constraint_segment)
                    for row in cursor.fetchall():
                        map = {}
                        for i, desc in enumerate(cursor.cursor.cursor.description):
                            map[desc[0]] = row[i]
                        segments.append(map)
                field_list = []
                for field in segments:
                    field_list.append('"' + field['FIELD_NAME'].strip() + '"')
                create_string += " unique(" + ','.join(str(x)
                                                       for x in field_list) + ")"
            try:
                editor.execute(create_string)
            except Exception as e:
                print(e)
        try:
            editor.execute("delete from django$constraint_segment")
            editor.execute("delete from django$constraint")
        except Exception as e:
            print(e)

    def table_exists(self, table_name):
        """
        Checks whether a table with a specified name exists.

        Args:
            table_name (str): Table name for checking
        """
        sql = """
            select null from rdb$relations where rdb$system_flag=0 and rdb$view_blr is null and rdb$relation_name='%s'
        """ % str(table_name).upper()
        value = None
        with self.cursor() as cursor:
            cursor.execute(sql)
            value = cursor.fetchone()
        return True if value else False

    def get_drop_constraints(self, query):
        """
        Returns objects that should be dropped when restrictions are disabled.

        Args:
            query (str): SQL query to execute

        .. important::

           The server does not support explicit disabling of restrictions,
           therefore, implementation is made using additional tables.

        Note:
           If there is an error when execute query an exception is thrown.
        """
        value = []
        with self.cursor() as cursor:
            cursor.execute(query)
            for row in cursor.fetchall():
                map = {}
                for i, desc in enumerate(cursor.cursor.cursor.description):
                    map[desc[0]] = row[i]
                value.append(map)
        return value

    # #### Backend-specific transaction management methods #####

    def _set_autocommit(self, autocommit):
        """
        Backend-specific implementation to enable or disable autocommit.

        FDB doesn't support auto-commit feature directly, but developers may
        achieve the similar result using explicit transaction start, taking
        advantage of default_action and its default value (commit).
        See:
        http://www.wikibasesql.org/file/documentation/drivers_documentation/python/fdb/usage-guide.html#auto-commit

        Pay attention at _close() method below
        """
        pass

    # #### Backend-specific wrappers for PEP-249 connection methods #####

    def _close(self):
        if self.connection is not None:
            with self.wrap_database_errors:
                if self.autocommit is True:
                    self.connection.commit()
                return self.connection.close()

    # #### Connection termination handling #####

    def is_usable(self):
        """
        Tests if the database connection is usable.
        This function may assume that self.connection is not None.
        """
        try:
            cur = self.connection.cursor()
            cur.execute('SELECT 1 FROM RDB$DATABASE')
        except DatabaseError:
            return False
        else:
            return True

    @cached_property
    def server_version(self):
        """
        Access method for engine_version property.
        engine_version return a full version in string format
        (ie: 'WI-V6.3.5.4926 Wikibase 1.5' )
        """
        if not self._server_version:
            if not self.connection:
                self.cursor()
            self._server_version = self.connection.db_info(
                WbDatabase._MEDIAWIKI_VERSION)
        return self._server_version


class WikibaseCursorWrapper(object):
    """
    Django uses "format" style placeholders, but wikibase uses "qmark" style.
    This fixes it -- but note that if you want to use a literal "%s" in a query,
    you'll need to use "%%s".
    """
    codes_for_integrityerror = (-803, -625, -530)

    def __init__(self, cursor, encoding):
        self.cursor = cursor
        self.encoding = encoding

    def execute(self, cmd, params=None):
        if params is None:
            params = []
        try:
            c = self.compile_cmd(cmd, len(params))
            return self.cursor.execute(c, params)
        except WbDatabase.IntegrityError as e:
            reraise(utils.IntegrityError, utils.IntegrityError(
                *self.error_info(e, cmd, params)), sys.exc_info()[2])
        except WbDatabase.DatabaseError as e:
            # Map some error codes to IntegrityError, since they seem to be
            # misclassified and Django would prefer the more logical place.
            # fdb: raise exception as tuple with (error_msg, sqlcode, error_code)
            code = self.get_sql_code(e)
            if code in self.codes_for_integrityerror:
                reraise(utils.IntegrityError, utils.IntegrityError(
                    *self.error_info(e, cmd, params)), sys.exc_info()[2])
            raise

    def executemany(self, cmd, param_list):
        try:
            q = self.compile_cmd(cmd, len(param_list[0]))
            return self.cursor.executemany(q, param_list)
        except WbDatabase.IntegrityError as e:
            reraise(utils.IntegrityError, utils.IntegrityError(
                *self.error_info(e, cmd, param_list[0])), sys.exc_info()[2])
        except WbDatabase.DatabaseError as e:
            # Map some error codes to IntegrityError, since they seem to be
            # misclassified and Django would prefer the more logical place.
            # fdb: raise exception as tuple with (error_msg, sqlcode, error_code)
            code = self.get_sql_code(e)
            if code in self.codes_for_integrityerror:
                reraise(utils.IntegrityError, utils.IntegrityError(
                    *self.error_info(e, cmd, param_list[0])), sys.exc_info()[2])
            raise

    def compile_cmd(self, cmd, num_params):
        """
            Check and compile
        """
        # if num_params == 0:
        #     return smart_str(query, self.encoding)
        # return smart_str(query % tuple("?" * num_params), self.encoding)
        return cmd

    def get_sql_code(self, e):
        try:
            sql_code = e.args[1]
        except IndexError:
            sql_code = None
        return sql_code

    def error_info(self, e, q, p):
        # fdb: raise exception as tuple with (error_msg, sqlcode, error_code)
        # just when it uses exception_from_status function. Ticket #44.
        try:
            error_msg = e.args[0]
        except IndexError:
            error_msg = ''

        try:
            sql_code = e.args[1]
        except IndexError:
            sql_code = None

        try:
            error_code = e.args[2]
        except IndexError:
            error_code = None

        if q:
            sql_text = q % tuple(p)
        else:
            sql_text = q
        return tuple([error_msg, sql_code, error_code, {'sql': sql_text, 'params': p}])

    def __getattr__(self, attr):
        if attr in self.__dict__:
            return self.__dict__[attr]
        else:
            return getattr(self.cursor, attr)

    def __iter__(self):
        return iter(self.cursor)



import datetime
from unittest import skipUnless

import django

from copy import deepcopy

from django.core.exceptions import FieldError, MultipleObjectsReturned
from django.db import models, transaction
from django.db.utils import IntegrityError
from django.test import TestCase, ignore_warnings
from six import assertRaisesRegex, text_type
from django.utils.translation import ugettext_lazy

if django.VERSION[0] < 2:
    # if django.version < 2.0
    from django.utils.deprecation import RemovedInDjango20Warning
else:
    RemovedInDjango21Warning = None


from tests.test_main.many_to_one.models import (
    Article, Category, Child, City, District, First, Parent, Record, Relation,
    Reporter, School, Student, Third, ToFieldChild,
)


class ManyToOneTests(TestCase):
    def setUp(self):
        # Create a few Reporters.
        self.r = Reporter(first_name='John', last_name='Smith', email='john@example.com')
        self.r.save()
        self.r2 = Reporter(first_name='Paul', last_name='Jones', email='paul@example.com')
        self.r2.save()
        # Create an Article.
        self.a = Article(headline="This is a test", pub_date=datetime.date(2005, 7, 27), reporter=self.r)
        self.a.save()

    def test_get(self):
        # Article objects have access to their related Reporter objects.
        r = self.a.reporter
        self.assertEqual(r.id, self.r.id)
        # These are strings instead of unicode strings because that's what was used in
        # the creation of this reporter (and we haven't refreshed the data from the
        # database, which always returns unicode strings).
        self.assertEqual((r.first_name, self.r.last_name), ('John', 'Smith'))

    def test_create(self):
        # You can also instantiate an Article by passing the Reporter's ID
        # instead of a Reporter object.
        a3 = Article(headline="Third article", pub_date=datetime.date(2005, 7, 27), reporter_id=self.r.id)
        a3.save()
        self.assertEqual(a3.reporter.id, self.r.id)

        # Similarly, the reporter ID can be a string.
        a4 = Article(headline="Fourth article", pub_date=datetime.date(2005, 7, 27), reporter_id=str(self.r.id))
        a4.save()
        self.assertEqual(repr(a4.reporter), "<Reporter: John Smith>")

    def test_add(self):
        # Create an Article via the Reporter object.
        new_article = self.r.article_set.create(headline="John's second story", pub_date=datetime.date(2005, 7, 29))
        self.assertEqual(repr(new_article), "<Article: John's second story>")
        self.assertEqual(new_article.reporter.id, self.r.id)

        # Create a new article, and add it to the article set.
        new_article2 = Article(headline="Paul's story", pub_date=datetime.date(2006, 1, 17))
        msg = "<Article: Paul's story> instance isn't saved. Use bulk=False or save the object first."
        with self.assertRaisesMessage(ValueError, msg):
            self.r.article_set.add(new_article2)

        self.r.article_set.add(new_article2, bulk=False)
        self.assertEqual(new_article2.reporter.id, self.r.id)
        self.assertQuerysetEqual(
            self.r.article_set.all(),
            ["<Article: John's second story>", "<Article: Paul's story>", "<Article: This is a test>"]
        )

        # Add the same article to a different article set - check that it moves.
        self.r2.article_set.add(new_article2)
        self.assertEqual(new_article2.reporter.id, self.r2.id)
        self.assertQuerysetEqual(self.r2.article_set.all(), ["<Article: Paul's story>"])

        # Adding an object of the wrong type raises TypeError.
        with transaction.atomic():
            with assertRaisesRegex(self, TypeError,
                                       "'Article' instance expected, got <Reporter.*"):
                self.r.article_set.add(self.r2)
        self.assertQuerysetEqual(
            self.r.article_set.all(),
            ["<Article: John's second story>", "<Article: This is a test>"]
        )

    def test_set(self):
        new_article = self.r.article_set.create(headline="John's second story", pub_date=datetime.date(2005, 7, 29))
        new_article2 = self.r2.article_set.create(headline="Paul's story", pub_date=datetime.date(2006, 1, 17))

        # Assign the article to the reporter.
        new_article2.reporter = self.r
        new_article2.save()
        self.assertEqual(repr(new_article2.reporter), "<Reporter: John Smith>")
        self.assertEqual(new_article2.reporter.id, self.r.id)
        self.assertQuerysetEqual(self.r.article_set.all(), [
            "<Article: John's second story>",
            "<Article: Paul's story>",
            "<Article: This is a test>",
        ])
        self.assertQuerysetEqual(self.r2.article_set.all(), [])

        # Set the article back again.
        self.r2.article_set.set([new_article, new_article2])
        self.assertQuerysetEqual(self.r.article_set.all(), ["<Article: This is a test>"])
        self.assertQuerysetEqual(
            self.r2.article_set.all(),
            ["<Article: John's second story>", "<Article: Paul's story>"]
        )

        # Funny case - because the ForeignKey cannot be null,
        # existing members of the set must remain.
        self.r.article_set.set([new_article])
        self.assertQuerysetEqual(
            self.r.article_set.all(),
            ["<Article: John's second story>", "<Article: This is a test>"]
        )
        self.assertQuerysetEqual(self.r2.article_set.all(), ["<Article: Paul's story>"])

    def test_reverse_assignment_deprecation(self):
        msg = (
            "Direct assignment to the reverse side of a related set is "
            "deprecated due to the implicit save() that happens. Use "
            "article_set.set() instead."
        )
        with self.assertRaisesMessage(RemovedInDjango20Warning, msg):
            self.r2.article_set = []

    def test_assign(self):
        new_article = self.r.article_set.create(headline="John's second story", pub_date=datetime.date(2005, 7, 29))
        new_article2 = self.r2.article_set.create(headline="Paul's story", pub_date=datetime.date(2006, 1, 17))

        # Assign the article to the reporter directly using the descriptor.
        new_article2.reporter = self.r
        new_article2.save()
        self.assertEqual(repr(new_article2.reporter), "<Reporter: John Smith>")
        self.assertEqual(new_article2.reporter.id, self.r.id)
        self.assertQuerysetEqual(self.r.article_set.all(), [
            "<Article: John's second story>",
            "<Article: Paul's story>",
            "<Article: This is a test>",
        ])
        self.assertQuerysetEqual(self.r2.article_set.all(), [])

        # Set the article back again using set() method.
        self.r2.article_set.set([new_article, new_article2])
        self.assertQuerysetEqual(self.r.article_set.all(), ["<Article: This is a test>"])
        self.assertQuerysetEqual(
            self.r2.article_set.all(),
            ["<Article: John's second story>", "<Article: Paul's story>"]
        )

        # Because the ForeignKey cannot be null, existing members of the set
        # must remain.
        self.r.article_set.set([new_article])
        self.assertQuerysetEqual(
            self.r.article_set.all(),
            ["<Article: John's second story>", "<Article: This is a test>"]
        )
        self.assertQuerysetEqual(self.r2.article_set.all(), ["<Article: Paul's story>"])
        # Reporter cannot be null - there should not be a clear or remove method
        self.assertFalse(hasattr(self.r2.article_set, 'remove'))
        self.assertFalse(hasattr(self.r2.article_set, 'clear'))

    def test_selects(self):
        self.r.article_set.create(headline="John's second story", pub_date=datetime.date(2005, 7, 29))
        self.r2.article_set.create(headline="Paul's story", pub_date=datetime.date(2006, 1, 17))
        # Reporter objects have access to their related Article objects.
        self.assertQuerysetEqual(self.r.article_set.all(), [
            "<Article: John's second story>",
            "<Article: This is a test>",
        ])
        self.assertQuerysetEqual(self.r.article_set.filter(headline__startswith='This'), ["<Article: This is a test>"])
        self.assertEqual(self.r.article_set.count(), 2)
        self.assertEqual(self.r2.article_set.count(), 1)
        # Get articles by id
        self.assertQuerysetEqual(Article.objects.filter(id__exact=self.a.id), ["<Article: This is a test>"])
        self.assertQuerysetEqual(Article.objects.filter(pk=self.a.id), ["<Article: This is a test>"])
        # Query on an article property
        self.assertQuerysetEqual(Article.objects.filter(headline__startswith='This'), ["<Article: This is a test>"])
        # The API automatically follows relationships as far as you need.
        # Use double underscores to separate relationships.
        # This works as many levels deep as you want. There's no limit.
        # Find all Articles for any Reporter whose first name is "John".
        self.assertQuerysetEqual(
            Article.objects.filter(reporter__first_name__exact='John'),
            ["<Article: John's second story>", "<Article: This is a test>"]
        )
        # Check that implied __exact also works
        self.assertQuerysetEqual(
            Article.objects.filter(reporter__first_name='John'),
            ["<Article: John's second story>", "<Article: This is a test>"]
        )
        # Query twice over the related field.
        self.assertQuerysetEqual(
            Article.objects.filter(reporter__first_name__exact='John', reporter__last_name__exact='Smith'),
            ["<Article: John's second story>", "<Article: This is a test>"]
        )
        # The underlying query only makes one join when a related table is referenced twice.
        queryset = Article.objects.filter(reporter__first_name__exact='John', reporter__last_name__exact='Smith')
        self.assertNumQueries(1, list, queryset)
        self.assertEqual(queryset.query.get_compiler(queryset.db).as_sql()[0].count('INNER JOIN'), 1)

        # The automatically joined table has a predictable name.
        self.assertQuerysetEqual(
            Article.objects.filter(reporter__first_name__exact='John').extra(
                where=["many_to_one_reporter.last_name='Smith'"]),
            ["<Article: John's second story>", "<Article: This is a test>"]
        )
        # ... and should work fine with the unicode that comes out of forms.Form.cleaned_data
        self.assertQuerysetEqual(
            (Article.objects
                .filter(reporter__first_name__exact='John')
                .extra(where=["many_to_one_reporter.last_name='%s'" % 'Smith'])),
            ["<Article: John's second story>", "<Article: This is a test>"]
        )
        # Find all Articles for a Reporter.
        # Use direct ID check, pk check, and object comparison
        self.assertQuerysetEqual(
            Article.objects.filter(reporter__id__exact=self.r.id),
            [
                "<Article: John's second story>",
                "<Article: This is a test>",
            ])
        self.assertQuerysetEqual(
            Article.objects.filter(reporter__pk=self.r.id),
            [
                "<Article: John's second story>",
                "<Article: This is a test>",
            ])
        self.assertQuerysetEqual(
            Article.objects.filter(reporter=self.r.id),
            [
                "<Article: John's second story>",
                "<Article: This is a test>",
            ])
        self.assertQuerysetEqual(
            Article.objects.filter(reporter=self.r),
            [
                "<Article: John's second story>",
                "<Article: This is a test>",
            ])
        self.assertQuerysetEqual(
            Article.objects.filter(reporter__in=[self.r.id, self.r2.id]).distinct(),
            [
                "<Article: John's second story>",
                "<Article: Paul's story>",
                "<Article: This is a test>",
            ])
        self.assertQuerysetEqual(
            Article.objects.filter(reporter__in=[self.r, self.r2]).distinct(),
            [
                "<Article: John's second story>",
                "<Article: Paul's story>",
                "<Article: This is a test>",
            ])
        # You can also use a queryset instead of a literal list of instances.
        # The queryset must be reduced to a list of values using values(),
        # then converted into a query
        self.assertQuerysetEqual(
            Article.objects.filter(
                reporter__in=Reporter.objects.filter(first_name='John').values('pk').query
            ).distinct(),
            [
                "<Article: John's second story>",
                "<Article: This is a test>",
            ])

    def test_reverse_selects(self):
        a3 = Article.objects.create(
            headline="Third article",
            pub_date=datetime.date(2005, 7, 27),
            reporter_id=self.r.id,
        )
        Article.objects.create(
            headline="Fourth article",
            pub_date=datetime.date(2005, 7, 27),
            reporter_id=self.r.id,
        )
        john_smith = ["<Reporter: John Smith>"]
        # Reporters can be queried
        self.assertQuerysetEqual(Reporter.objects.filter(id__exact=self.r.id), john_smith)
        self.assertQuerysetEqual(Reporter.objects.filter(pk=self.r.id), john_smith)
        self.assertQuerysetEqual(Reporter.objects.filter(first_name__startswith='John'), john_smith)
        # Reporters can query in opposite direction of ForeignKey definition
        self.assertQuerysetEqual(Reporter.objects.filter(article__id__exact=self.a.id), john_smith)
        self.assertQuerysetEqual(Reporter.objects.filter(article__pk=self.a.id), john_smith)
        self.assertQuerysetEqual(Reporter.objects.filter(article=self.a.id), john_smith)
        self.assertQuerysetEqual(Reporter.objects.filter(article=self.a), john_smith)
        self.assertQuerysetEqual(Reporter.objects.filter(article__in=[self.a.id, a3.id]).distinct(), john_smith)
        self.assertQuerysetEqual(Reporter.objects.filter(article__in=[self.a.id, a3]).distinct(), john_smith)
        self.assertQuerysetEqual(Reporter.objects.filter(article__in=[self.a, a3]).distinct(), john_smith)
        self.assertQuerysetEqual(
            Reporter.objects.filter(article__headline__startswith='T'),
            ["<Reporter: John Smith>", "<Reporter: John Smith>"],
            ordered=False
        )
        self.assertQuerysetEqual(Reporter.objects.filter(article__headline__startswith='T').distinct(), john_smith)

        # Counting in the opposite direction works in conjunction with distinct()
        self.assertEqual(Reporter.objects.filter(article__headline__startswith='T').count(), 2)
        self.assertEqual(Reporter.objects.filter(article__headline__startswith='T').distinct().count(), 1)

        # Queries can go round in circles.
        self.assertQuerysetEqual(
            Reporter.objects.filter(article__reporter__first_name__startswith='John'),
            [
                "<Reporter: John Smith>",
                "<Reporter: John Smith>",
                "<Reporter: John Smith>",
            ],
            ordered=False
        )
        self.assertQuerysetEqual(
            Reporter.objects.filter(article__reporter__first_name__startswith='John').distinct(),
            john_smith
        )
        self.assertQuerysetEqual(Reporter.objects.filter(article__reporter__exact=self.r).distinct(), john_smith)

        # Check that implied __exact also works.
        self.assertQuerysetEqual(Reporter.objects.filter(article__reporter=self.r).distinct(), john_smith)

        # It's possible to use values() calls across many-to-one relations.
        # (Note, too, that we clear the ordering here so as not to drag the
        # 'headline' field into the columns being used to determine uniqueness)
        d = {'reporter__first_name': 'John', 'reporter__last_name': 'Smith'}
        qs = Article.objects.filter(
            reporter=self.r,
        ).distinct().order_by().values('reporter__first_name', 'reporter__last_name')
        self.assertEqual([d], list(qs))

    def test_select_related(self):
        # Check that Article.objects.select_related().dates() works properly when
        # there are multiple Articles with the same date but different foreign-key
        # objects (Reporters).
        r1 = Reporter.objects.create(first_name='Mike', last_name='Royko', email='royko@suntimes.com')
        r2 = Reporter.objects.create(first_name='John', last_name='Kass', email='jkass@tribune.com')
        Article.objects.create(headline='First', pub_date=datetime.date(1980, 4, 23), reporter=r1)
        Article.objects.create(headline='Second', pub_date=datetime.date(1980, 4, 23), reporter=r2)
        self.assertEqual(
            list(Article.objects.select_related().dates('pub_date', 'day')),
            [datetime.date(1980, 4, 23), datetime.date(2005, 7, 27)]
        )
        self.assertEqual(
            list(Article.objects.select_related().dates('pub_date', 'month')),
            [datetime.date(1980, 4, 1), datetime.date(2005, 7, 1)]
        )
        self.assertEqual(
            list(Article.objects.select_related().dates('pub_date', 'year')),
            [datetime.date(1980, 1, 1), datetime.date(2005, 1, 1)]
        )

    def test_delete(self):
        self.r.article_set.create(headline="John's second story", pub_date=datetime.date(2005, 7, 29))
        self.r2.article_set.create(headline="Paul's story", pub_date=datetime.date(2006, 1, 17))
        Article.objects.create(headline="Third article", pub_date=datetime.date(2005, 7, 27), reporter_id=self.r.id)
        Article.objects.create(
            headline="Fourth article",
            pub_date=datetime.date(2005, 7, 27),
            reporter_id=str(self.r.id),
        )
        # If you delete a reporter, his articles will be deleted.
        self.assertQuerysetEqual(
            Article.objects.all(),
            [
                "<Article: Fourth article>",
                "<Article: John's second story>",
                "<Article: Paul's story>",
                "<Article: Third article>",
                "<Article: This is a test>",
            ]
        )
        self.assertQuerysetEqual(
            Reporter.objects.order_by('first_name'),
            ["<Reporter: John Smith>", "<Reporter: Paul Jones>"]
        )
        self.r2.delete()
        self.assertQuerysetEqual(
            Article.objects.all(),
            [
                "<Article: Fourth article>",
                "<Article: John's second story>",
                "<Article: Third article>",
                "<Article: This is a test>",
            ]
        )
        self.assertQuerysetEqual(Reporter.objects.order_by('first_name'), ["<Reporter: John Smith>"])
        # You can delete using a JOIN in the query.
        Reporter.objects.filter(article__headline__startswith='This').delete()
        self.assertQuerysetEqual(Reporter.objects.all(), [])
        self.assertQuerysetEqual(Article.objects.all(), [])

    def test_explicit_fk(self):
        # Create a new Article with get_or_create using an explicit value
        # for a ForeignKey.
        a2, created = Article.objects.get_or_create(
            headline="John's second test",
            pub_date=datetime.date(2011, 5, 7),
            reporter_id=self.r.id,
        )
        self.assertTrue(created)
        self.assertEqual(a2.reporter.id, self.r.id)

        # You can specify filters containing the explicit FK value.
        self.assertQuerysetEqual(
            Article.objects.filter(reporter_id__exact=self.r.id),
            ["<Article: John's second test>", "<Article: This is a test>"]
        )

        # Create an Article by Paul for the same date.
        a3 = Article.objects.create(
            headline="Paul's commentary",
            pub_date=datetime.date(2011, 5, 7),
            reporter_id=self.r2.id,
        )
        self.assertEqual(a3.reporter.id, self.r2.id)

        # Get should respect explicit foreign keys as well.
        with self.assertRaises(MultipleObjectsReturned):
            Article.objects.get(reporter_id=self.r.id)
        self.assertEqual(
            repr(a3),
            repr(Article.objects.get(reporter_id=self.r2.id, pub_date=datetime.date(2011, 5, 7)))
        )

    def test_deepcopy_and_circular_references(self):
        # Regression for #12876 -- Model methods that include queries that
        # recursive don't cause recursion depth problems under deepcopy.
        self.r.cached_query = Article.objects.filter(reporter=self.r)
        self.assertEqual(repr(deepcopy(self.r)), "<Reporter: John Smith>")

    def test_manager_class_caching(self):
        r1 = Reporter.objects.create(first_name='Mike')
        r2 = Reporter.objects.create(first_name='John')

        # Same twice
        self.assertIs(r1.article_set.__class__, r1.article_set.__class__)

        # Same as each other
        self.assertIs(r1.article_set.__class__, r2.article_set.__class__)

    def test_create_relation_with_ugettext_lazy(self):
        reporter = Reporter.objects.create(first_name='John', last_name='Smith', email='john.smith@example.com')
        lazy = ugettext_lazy('test')
        reporter.article_set.create(headline=lazy, pub_date=datetime.date(2011, 6, 10))
        notlazy = text_type(lazy)
        article = reporter.article_set.get()
        self.assertEqual(article.headline, notlazy)

    def test_values_list_exception(self):
        expected_message = "Cannot resolve keyword 'notafield' into field. Choices are: %s"
        reporter_fields = ', '.join(sorted(f.name for f in Reporter._meta.get_fields()))
        with self.assertRaisesMessage(FieldError, expected_message % reporter_fields):
            Article.objects.values_list('reporter__notafield')
        article_fields = ', '.join(['EXTRA'] + sorted(f.name for f in Article._meta.get_fields()))
        with self.assertRaisesMessage(FieldError, expected_message % article_fields):
            Article.objects.extra(select={'EXTRA': 'EXTRA_SELECT'}).values_list('notafield')

    def test_fk_assignment_and_related_object_cache(self):
        # Tests of ForeignKey assignment and the related-object cache (see #6886).

        p = Parent.objects.create(name="Parent")
        c = Child.objects.create(name="Child", parent=p)

        # Look up the object again so that we get a "fresh" object.
        c = Child.objects.get(name="Child")
        p = c.parent

        # Accessing the related object again returns the exactly same object.
        self.assertIs(c.parent, p)

        # But if we kill the cache, we get a new object.
        del c._parent_cache
        self.assertIsNot(c.parent, p)

        # Assigning a new object results in that object getting cached immediately.
        p2 = Parent.objects.create(name="Parent 2")
        c.parent = p2
        self.assertIs(c.parent, p2)

        # Assigning None succeeds if field is null=True.
        p.bestchild = None
        self.assertIsNone(p.bestchild)

        # bestchild should still be None after saving.
        p.save()
        self.assertIsNone(p.bestchild)

        # bestchild should still be None after fetching the object again.
        p = Parent.objects.get(name="Parent")
        self.assertIsNone(p.bestchild)

        # Assigning None will not fail: Child.parent is null=False.
        setattr(c, "parent", None)

        # You also can't assign an object of the wrong type here
        with self.assertRaises(ValueError):
            setattr(c, "parent", First(id=1, second=1))

        # You can assign None to Child.parent during object creation.
        Child(name='xyzzy', parent=None)

        # But when trying to save a Child with parent=None, the database will
        # raise IntegrityError.
        with self.assertRaises(IntegrityError), transaction.atomic():
            Child.objects.create(name='xyzzy', parent=None)

        # Creation using keyword argument should cache the related object.
        p = Parent.objects.get(name="Parent")
        c = Child(parent=p)
        self.assertIs(c.parent, p)

        # Creation using keyword argument and unsaved related instance (#8070).
        p = Parent()
        msg = "save() prohibited to prevent data loss due to unsaved related object 'parent'."
        with self.assertRaisesMessage(ValueError, msg):
            Child.objects.create(parent=p)

        msg = "save() prohibited to prevent data loss due to unsaved related object 'parent'."
        with self.assertRaisesMessage(ValueError, msg):
            ToFieldChild.objects.create(parent=p)

        # Creation using attname keyword argument and an id will cause the
        # related object to be fetched.
        p = Parent.objects.get(name="Parent")
        c = Child(parent_id=p.id)
        self.assertIsNot(c.parent, p)
        self.assertEqual(c.parent, p)

    def test_fk_to_bigautofield(self):
        ch = City.objects.create(name='Chicago')
        District.objects.create(city=ch, name='Far South')
        District.objects.create(city=ch, name='North')

        ny = City.objects.create(name='New York', id=2 ** 33)
        District.objects.create(city=ny, name='Brooklyn')
        District.objects.create(city=ny, name='Manhattan')

    def test_multiple_foreignkeys(self):
        # Test of multiple ForeignKeys to the same model (bug #7125).
        c1 = Category.objects.create(name='First')
        c2 = Category.objects.create(name='Second')
        c3 = Category.objects.create(name='Third')
        r1 = Record.objects.create(category=c1)
        r2 = Record.objects.create(category=c1)
        r3 = Record.objects.create(category=c2)
        r4 = Record.objects.create(category=c2)
        r5 = Record.objects.create(category=c3)
        Relation.objects.create(left=r1, right=r2)
        Relation.objects.create(left=r3, right=r4)
        Relation.objects.create(left=r1, right=r3)
        Relation.objects.create(left=r5, right=r2)
        Relation.objects.create(left=r3, right=r2)

        q1 = Relation.objects.filter(left__category__name__in=['First'], right__category__name__in=['Second'])
        self.assertQuerysetEqual(q1, ["<Relation: First - Second>"])

        q2 = Category.objects.filter(record__left_set__right__category__name='Second').order_by('name')
        self.assertQuerysetEqual(q2, ["<Category: First>", "<Category: Second>"])

        p = Parent.objects.create(name="Parent")
        c = Child.objects.create(name="Child", parent=p)
        with self.assertRaises(ValueError):
            Child.objects.create(name="Grandchild", parent=c)

    def test_fk_instantiation_outside_model(self):
        # Regression for #12190 -- Should be able to instantiate a FK outside
        # of a model, and interrogate its related field.
        cat = models.ForeignKey(Category, models.CASCADE)
        self.assertEqual('id', cat.remote_field.get_related_field().name)

    def test_relation_unsaved(self):
        # Test that the <field>_set manager does not join on Null value fields (#17541)
        Third.objects.create(name='Third 1')
        Third.objects.create(name='Third 2')
        th = Third(name="testing")
        # The object isn't saved an thus the relation field is null - we won't even
        # execute a query in this case.
        with self.assertNumQueries(0):
            self.assertEqual(th.child_set.count(), 0)
        th.save()
        # Now the model is saved, so we will need to execute an query.
        with self.assertNumQueries(1):
            self.assertEqual(th.child_set.count(), 0)

    @skipUnless(RemovedInDjango20Warning == None, "RemovedInDjango20Warning is not defined")
    @ignore_warnings(category=RemovedInDjango20Warning)  # for use_for_related_fields deprecation
    def test_related_object(self):
        public_school = School.objects.create(is_public=True)
        public_student = Student.objects.create(school=public_school)

        private_school = School.objects.create(is_public=False)
        private_student = Student.objects.create(school=private_school)

        # Only one school is available via all() due to the custom default manager.
        self.assertQuerysetEqual(School.objects.all(), ["<School: School object>"])

        self.assertEqual(public_student.school, public_school)

        # Make sure the base manager is used so that an student can still access
        # its related school even if the default manager doesn't normally
        # allow it.
        self.assertEqual(private_student.school, private_school)

        # If the manager is marked "use_for_related_fields", it'll get used instead
        # of the "bare" queryset. Usually you'd define this as a property on the class,
        # but this approximates that in a way that's easier in tests.
        School._default_manager.use_for_related_fields = True
        try:
            private_student = Student.objects.get(pk=private_student.pk)
            with self.assertRaises(School.DoesNotExist):
                private_student.school
        finally:
            School._default_manager.use_for_related_fields = False

        School._meta.base_manager_name = 'objects'
        School._meta._expire_cache()
        try:
            private_student = Student.objects.get(pk=private_student.pk)
            with self.assertRaises(School.DoesNotExist):
                private_student.school
        finally:
            School._meta.base_manager_name = None
            School._meta._expire_cache()

    def test_hasattr_related_object(self):
        # The exception raised on attribute access when a related object
        # doesn't exist should be an instance of a subclass of `AttributeError`
        # refs #21563
        self.assertFalse(hasattr(Article(), 'reporter'))
